
/*                   MIXNLIN (Macro version 3.1)                  */
/*  New Features: 1. SS IRGLS Based on Shrinkage Estimators       */
/*                2. Includes a VARPARMS statement                */
/*                3. Includes an OUTPRED statement outputting     */
/*                   observed and predicted values.               */
/*                4. Includes an ESTIMATE statement               */
/*                5. Change PNMEM F-TEST to MANOVA F-TEST         */
/*                6. Put a CYCLE Option in for EGLS               */
/*                7. Corrected STDERR statement (it was COVAR=)   */
/*                8. New Concordance Corr based on Omega          */
/*                9. New RESTRICT Statement for prop. errors      */

/*  Procedure MIXNLIN is a SAS based Macro/IML procedure written  */
/*  explicitly for applications involving nonlinear mixed_effects */
/*  models.  MIXNLIN was written and tested by:                   */
/*                 Edward F. Vonesh, Ph.D.                        */
/*                 Applied Statistics Center                      */
/*                 Baxter Healthcare Corporation                  */
/*                 Route 120 & Wilson Rd.                         */
/*                 Round Lake, IL 60073 USA                       */
/*  Presented below are the IML Statements which must be compiled */
/*  and saved under an appropriate file (see user manual).  The   */
/*  LIBNAME and RESET statements below can be redefined for this  */
/*  purpose (again see user manual for details).                  */


/*                       LIBNAME statement below                    */
/* Use the LIBNAME here to define the HOST dir containing this file */
LIBNAME RMLIB '[EDWARD.MIXNLIN]';

/*  IML SUBROUTINES DEFINED HERE  */

PROC IML ;
/*                         RESET statement below                       */
/* Use the RESET statement here to define the IML library storage file */
RESET STORAGE=RMLIB.STOREMOD;





START GENMEM;      /* GENMEM = GENERALIZED MIXED-EFFECTS MODELS MODULE */
/* ** SET PARAMETERS ** */
SIGSQIND= (SIGSQ>0);         * (=1 IF SIGSQ>0, =0 IF NOT);
SIG_DELT=SIGSQIND+SIG82152;  * (=1 IF SIGSQ>=0, =0 IF NOT);
B082152 = REPEAT(0,TOTAL,P); * INITIALIZE B082152=(TOTAL X P)MATRIX OF ZEROES;
BI_HAT  = REPEAT(0,TOTAL,V); * INITIALIZE SUBJECT-SPECIFIC RANDOM EFFECTS = 0;
XI_HAT  = REPEAT(0,TOTAL,V); * INITIALIZE OLS RANDOM EFFECTS = 0;

/*  MODULE FOR COMPUTING DETERMINANT AND LOG(DETERMINANT) OF VI */
START LOGDET(AAAA) GLOBAL(LNDETERM,RI,BYVAR,FIRST);
 SCALE=MAX(VECDIAG(AAAA));
 IF SCALE<1 THEN SCALE=1;
 VI_SCALE=AAAA/SCALE;
 CHECK1=DET(VI_SCALE);
 IF CHECK1>0 THEN GO TO ENDSC2;
 DO III=1 TO 100 UNTIL(CHECK1 > 0);
  IF SCALE<1 THEN GO TO ENDSC1;
  ELSE SCALE=SCALE/10;
  VI_SCALE=AAAA/SCALE;
  CHECK1=DET(VI_SCALE);
 ENDSC1:
 END;
 ENDSC2:
 IF CHECK1=0 THEN DO;
 LNDETERM=.;
 ID=BYVAR[FIRST,];
 PRINT 'ERROR: AN INDIVIDUAL VARIANCE-COVARIANCE MATRIX IS SINGULAR',
       '       ERROR OCCURRED FOR INDIVIDUAL:' ID,,,
       'VARIANCE-COVARIANCE = ' AAAA;
 END;
 ELSE DO;
 LNDETERM = LNDETERM + ( RI*LOG(SCALE) + LOG(CHECK1) );
 END;
  RETURN (LNDETERM);
FINISH;

/*  MODULE FOR COMPUTING LOG(DETERMINANT) OF ANY MATRIX */
START LOG_DET(AAAA);
 SCALE=MAX(VECDIAG(AAAA));
 IF SCALE<1 THEN SCALE=1;
 VI_SCALE=AAAA/SCALE;
 CHECK1=DET(VI_SCALE);
 IF CHECK1>0 THEN GO TO ENDSC2;
 DO III=1 TO 100 UNTIL(CHECK1 > 0);
  IF SCALE<1 THEN GO TO ENDSC1;
  ELSE SCALE=SCALE/10;
  VI_SCALE=AAAA/SCALE;
  CHECK1=DET(VI_SCALE);
 ENDSC1:
 END;
 ENDSC2:
 IF CHECK1=0 THEN DO;
 _LOGDET_=LOG(.10**20);
 END;
 ELSE DO;
 _LOGDET_ = ( NCOL(AAAA)*LOG(SCALE) + LOG(CHECK1) );
 END;
  RETURN (_LOGDET_);
FINISH;


/*  NEW MODULE FOR COMPUTING INVERSE OF VI */
START INVERSE(AAA,PSI_INV,SIG_INV) GLOBAL(RANDOM,SIG82152,DI,ZI,V);
 IF (SIG82152=0)|(RANDOM=0)=1 THEN DO;
  INV_DI=INV(SIG_INV*DI);
  BBB = INV_DI - INV_DI*ZI*PSI_INV*INV( I(V)+ZI`*INV_DI*ZI*PSI_INV)*ZI`*INV_DI;
  END;
  ELSE DO;
  BBB = INV(ZI*PSI_INV*ZI`);
  END;
 RETURN(BBB);
FINISH;

/* NEWEST MODULE FOR CALCULATING PRED_SS UNDER SS OR PA APPROXIMATION */
START PRED_SS;
 B082152=BI_HAT*SELECT; /* THIS WILL EVALUATE F(BETA+bi) FOR PA AND SS MODELS */
 RUN MODEL;
 RUN RANDOM;
  PRED_SS=F;NCUM=0;
 DO J=1 TO N;
  RI=SAMSIZ[J,];
  FIRST=NCUM+1;LAST=NCUM+RI;
  PRED_SS[FIRST:LAST,] = PRED_SS[FIRST:LAST,] +
                         RANDOM*ZFIX[FIRST:LAST,]*SELECT2`*BI_HAT[FIRST,]`;
/* THE ABOVE EVALUATES OTHER LINEAR RANDOM EFFECTS GIVEN IN RANDOM STATEMENT */
  NCUM=NCUM+RI;
 END;
 B082152=S082152*B082152; /* RESET B082152, F, X, Z ETC. TO INTENDED VALUE */
 RUN MODEL;
 RUN RANDOM;
FINISH;

/* NEWEST MODULE FOR CALCULATING CONCORDANCE CORRELATIONS */
START CONCORR;
/* RHO_PA AND RHO_SS ARE CONCORDANCE CORRELATIONS FOR Y */
 RUN PRED_SS;
 YOBS=Y082152;
 YHAT=_PRED_;         * = f(B) FOR PA AND SS MODELS;
 YHAT_SS=PRED_SS;     * = f(B+bi) FOR PA AND SS MODELS;
  SS_OBS=YOBS`*(I(TOTAL)-(1/TOTAL)*J(TOTAL,1,1)*J(TOTAL,1,1)`)*YOBS;
  SS_HAT=YHAT`*(I(TOTAL)-(1/TOTAL)*J(TOTAL,1,1)*J(TOTAL,1,1)`)*YHAT;
  SS_HATI=YHAT_SS`*(I(TOTAL)-(1/TOTAL)*J(TOTAL,1,1)*J(TOTAL,1,1)`)*YHAT_SS;
  AVG_OBS=SUM(YOBS)/TOTAL;
  AVG_HAT=SUM(YHAT)/TOTAL;
  AVG_HATI=SUM(YHAT_SS)/TOTAL;
  SS_AVG=(AVG_OBS-AVG_HAT)**2;
  SS_AVGI=(AVG_OBS-AVG_HATI)**2;
  RHO_PA=1 - (YOBS-YHAT)`*(YOBS-YHAT)/(SS_OBS+SS_HAT+TOTAL*SS_AVG);
  RHO_PA=1 - (1-RHO_PA)*(TOTAL/(TOTAL-PQ));
  RHO_SS=1 - (YOBS-YHAT_SS)`*(YOBS-YHAT_SS)/(SS_OBS+SS_HATI+TOTAL*SS_AVGI);
  RHO_SS=1 - (1-RHO_SS)*(TOTAL/(TOTAL-PQ));
  RSQ_PA = 1 - (YOBS-YHAT)`*(YOBS-YHAT)/SS_OBS;
  RSQ_PA = 1 - (1-RSQ_PA)*(TOTAL/(TOTAL-PQ));
  RSQ_SS = 1 - (YOBS-YHAT_SS)`*(YOBS-YHAT_SS)/SS_OBS;
  RSQ_SS = 1 - (1-RSQ_SS)*(TOTAL/(TOTAL-PQ));

FINISH;


START AIC;         /* MODULE FOR COMPUTING AIC AND SBC CRITERIA OF FIT */
 RESID=Y082152-F;
  XWX=0;XWY=0;YWY=0;LNDETERM=0;NCUM=0;
 DO J=1 TO N;
  RI=SAMSIZ[J,];
  FIRST=NCUM+1;LAST=NCUM+RI;
  RESID_I=RESID[FIRST:LAST,];
  XI=X[FIRST:LAST,];
  ZI=RANDOM*Z[FIRST:LAST,];
  BETAI_I=B082152[FIRST,]*SELECT`*S082152;
  RESID_I=RESID_I+ZI*BETAI_I`;
  DI=_WEIGHT_[FIRST:LAST,];
  DI=INV(DIAG(DI));
  VI = ZI*PSI*ZI` + SIGPOOL*DI;
  LNDETERM=LOGDET(VI);
  WI=INVERSE(VI,PSI,SIGPOOL);
  XWX = XWX + XI`*WI*XI;
  XWY = XWY + XI`*WI*RESID_I;
  YWY = YWY + RESID_I`*WI*RESID_I;
  NCUM=NCUM+RI;
 END;
  Q_SS=YWY;
  OBJECT=LNDETERM+Q_SS;

* Following statement tries to use correct components of XWX when RESTRICT
  statement is in effect. ;
SUBXWX=XWX[LOC(VECDIAG(XWX)>0),LOC(VECDIAG(XWX)>0)];
CONST=LOG_DET(SUBXWX);

/* LOGLIKE2 = -2 LOG LIKELIHOOD WITH MLE_IND AN INDICATOR FOR MLE VS. REML */
  LOGLIKE2=(TOTAL-(1-MLE_IND)*PQ)*LOG(2*3.14159)
            + OBJECT + (1-MLE_IND)*CONST;
  QQ = RANDOM*NROW(PSI)*(NROW(PSI)+1)/2 + (1-SIG_DELT);

/* PQ = NUMBER FIXED EFFECTS, QQ = NUMBER OF COVARIANCE PARAMETERS */
AIC = -.5*LOGLIKE2 - (PQ+QQ);
SBC = -.5*LOGLIKE2 - .5*(PQ+QQ)*LOG(TOTAL-(1-MLE_IND)*PQ);
FINISH;


START GRID;        /* GRID MODULE PERFORMS GRID SEARCH */
 STARTS=NROW(ALPHA0);
 GRID=ALPHA0||J(STARTS,1,0);
 ENDS=NCOL(GRID);

DO I=1 TO STARTS;
ALPH_EST=ALPHA0[I,]`;
 RUN MODEL;
 RESID=Y082152-F;
 Q_SS0=0;NCUM=0;
 DO J=1 TO N;
  RI=SAMSIZ[J,];
  FIRST=NCUM+1;LAST=NCUM+RI;
  RESID_I=RESID[FIRST:LAST,];
  DI=_WEIGHT_[FIRST:LAST,];
  DI=DIAG(DI);
  Q_SS0 = Q_SS0 + RESID_I`*DI*RESID_I;
  NCUM=NCUM+RI;
 END;
GRID[I,ENDS]=Q_SS0;
END;
MIN_Q_SS=MIN(GRID[,ENDS]);
ALPHA0=GRID[LOC(RANK(GRID[,ENDS])=1),(1:(ENDS-1))];
ALPHA0=ALPHA0`;

IF BEST>0 THEN DO;
  BEST=MIN(BEST,STARTS);
  BESTGRID=GRID;
  GRID[RANK(GRID[,ENDS]),]=BESTGRID;
  NEWGRID=GRID[1:BEST,];
RESET NONAME;
GRIDN=ALPHAN`||{"RESIDUAL SS"};
PRINT / 'MACRO PROCEDURE MIXNLIN',,
        'PRINT OF NONLINEAR LEAST SQUARES GRID SEARCH',,
        'DEPENDENT VARIABLE' NAMEY,,
        NEWGRID (|COLNAME=GRIDN|);
END;

FINISH;


START ESTIMATE;     /* DETERMINES UPDATE VALUE DELTA FOR LEAST SQUARES  */
 RUN MODEL;
 RESID=Y082152-F;
 BB082152 = B082152;
 XWX=0;XWY=0;YWY=0;Q_SS=0;NCUM=0;
 DO J=1 TO N;
  RI=SAMSIZ[J,];
  FIRST=NCUM+1;LAST=NCUM+RI;
  RESID_I=RESID[FIRST:LAST,];
  XI=X[FIRST:LAST,];
  ZI=RANDOM*Z[FIRST:LAST,];

*NEWEST FOR IRGLS;
  BETAI_I=B082152[FIRST,]*SELECT`*S082152;
  RESID_I=RESID_I+ZI*BETAI_I`;
  DI=_WEIGHT_[FIRST:LAST,];
  DI=INV(DIAG(DI));
  VI = ZI*PSI_EST*ZI` + SIG_EST*DI;
  WI=INVERSE(VI,PSI_EST,SIG_EST);
  XWX = XWX + XI`*WI*XI;
  XWY = XWY + XI`*WI*RESID_I;
  YWY = YWY + RESID_I`*WI*RESID_I;
  NCUM=NCUM+RI;
 END;
 Q_SS = YWY;
 OMEGA=GINV(XWX);
 DELTA=OMEGA*XWY;
FINISH;


/* DETERMINES UPDATED VALUE OF BI_GLS FOR (CYCLED EGLS)/(IRGLS) */
/* RUN THIS AFTER EACH ESTIMATE STATEMENT IN IRGLS */
/* AND BETWEEN CYCLES OF EGLS STATEMENT            */
START ESTIM_BI;
NCUM=0;
 DO J=1 TO N;
  RI=SAMSIZ[J,];
  FIRST=NCUM+1;LAST=NCUM+RI;
  RESID_I=RESID[FIRST:LAST,];
  XI=X[FIRST:LAST,];
  ZI=Z[FIRST:LAST,];
  BETAI_I=B082152[FIRST,]*SELECT`*S082152;
  RESID_I=RESID_I+ZI*BETAI_I`;
  DI=_WEIGHT_[FIRST:LAST,];
  DI=INV(DIAG(DI));
  VI = ZI*PSI_EST*ZI` + SIG_EST*DI;
  WI=INVERSE(VI,PSI_EST,SIG_EST);

/* FIRST BI_GLS IS BASED ON ASSUMING FULL RANK ZI FOR ALL SUBJECTS */
/* IT GIVES THE SAME ANSWER AS WHEN WE USE THE SECOND VERSION WITH */
/* GINV INSTEAD OF REGULAR INVERSE.                                */
/* BI_GLS = C082152*RANDOM*PSI_EST*ZI`*WI*(RESID_I-XI*DELTA) +     */
/*     (1-C082152)*RANDOM*INV(ZI`*DI*ZI)*ZI`*DI*(RESID_I-XI*DELTA);*/
  BI_GLS = C082152*RANDOM*PSI_EST*ZI`*WI*(RESID_I-XI*DELTA) +
           (1-C082152)*RANDOM*GINV(SQRT(DI)*ZI)*(SQRT(DI)*(RESID_I-XI*DELTA));
  BI_HAT[FIRST:LAST,] = REPEAT(BI_GLS`,RI,1);
  BB082152[FIRST:LAST,]=REPEAT(BI_GLS`*SELECT*S082152,RI,1);
  NCUM=NCUM+RI;
 END;
 B082152=BB082152;
FINISH;

START ROBUST;   /* CALCULATES ROBUST STANDARD ERRORS OF FIXED EFFECTS */
 _OMEGA_=OMEGA; * THIS RETAINS ORIGINAL OMEGA FOR MODULE OUTPUT;
 RUN MODEL;
 RESID=Y082152-F;
 XWX=0;XWY=0;YWY=0;NCUM=0;XWYYWX=0;
 DO J=1 TO N;
  RI=SAMSIZ[J,];
  FIRST=NCUM+1;LAST=NCUM+RI;
  RESID_I=RESID[FIRST:LAST,];
  XI=X[FIRST:LAST,];
  ZI=RANDOM*Z[FIRST:LAST,];
  BETAI_I=B082152[FIRST,]*SELECT`*S082152;
  RESID_I=RESID_I+ZI*BETAI_I`;
  DI=_WEIGHT_[FIRST:LAST,];
  DI=INV(DIAG(DI));
  VI = ZI*PSI_EST*ZI` + SIG_EST*DI;
  WI=INVERSE(VI,PSI_EST,SIG_EST);
  XWX = XWX + XI`*WI*XI;
  XWY = XWY + XI`*WI*RESID_I;
  YWY = YWY + RESID_I`*WI*RESID_I;
  XWYYWX = XWYYWX + XI`*WI*RESID_I*RESID_I`*WI*XI;
  NCUM=NCUM+RI;
 END;
 _XWYYWX_=XWYYWX[LOC(VECDIAG(XWYYWX)>0),LOC(VECDIAG(XWYYWX)>0)];
 OMEGA_A = OMEGA[LOC(VECDIAG(OMEGA)>0),LOC(VECDIAG(OMEGA)>0)];
 OMEGA_R = OMEGA_A*_XWYYWX_*OMEGA_A;  * THE REVISED SANDWICH ESTIMATOR *;

/* THE FOLLOWING IS SO WE DO NOT PERFORM LR TEST IF RESTRICT IS IN EFFECT */

 DETERR=
  "if overflow then do;
   OMEGACHK=1;
   CALL PUSH(DETERR); RESUME; END;";

  CALL PUSH(DETERR);
  overflow=1;
 OMEGACHK=(DET(OMEGA_A)=0)|(DET(OMEGA_R)=0);
  overflow=0;

 IF OMEGACHK=1 THEN GOTO OMEGARUN;
 SQOMEGA = GINV(ROOT(OMEGA_A));
 ASYVAR = SYMSQR(SQOMEGA`*OMEGA_R*SQOMEGA);
 ROBVAR = SYMSQR(I(PQ));
 DIFVAR = ASYVAR - ROBVAR;
 NVAR=NROW(ASYVAR);
  MSG1={" "};
  RHO_CVAR = 1 - (DIFVAR`*DIFVAR)/(ASYVAR`*ASYVAR + ROBVAR`*ROBVAR);
  LRTEST= N*( LOG_DET(OMEGA_A) - LOG_DET(OMEGA_R) +
           TRACE(OMEGA_R*GINV(OMEGA_A)) - PQ );
  DFLRTEST = .5*PQ*(PQ+1);
  P_LRTEST=1-PROBCHI(MIN(LRTEST,200),DFLRTEST);
  P_LRTEST=P_LRTEST<>.0001;
 OMEGARUN:
 IF OMEGACHK=1 THEN DO;
  msg1={"Singular Estimate of Omega or Overflow error detected: Test not done"};
  RHO_CVAR=.;LRTEST=.;DFLRTEST=.;P_LRTEST=.;
 END;
FINISH;

START SEARCH;      /* DEFINES Q_LAMDA USED IN HALF-STEP LEAST SQUARES  */
 RUN MODEL;
 *RUN RANDOM;
 RESID=Y082152-F;
 Q_LAMDA=0;NCUM=0;
 DO J=1 TO N;
  RI=SAMSIZ[J,];
  FIRST=NCUM+1;LAST=NCUM+RI;
  RESID_I=RESID[FIRST:LAST,];
  ZI=RANDOM*Z[FIRST:LAST,];

  BETAI_I=B082152[FIRST,]*SELECT`*S082152;
  RESID_I=RESID_I+ZI*BETAI_I`;

  DI=_WEIGHT_[FIRST:LAST,];
  DI=INV(DIAG(DI));
  VI = ZI*PSI_EST*ZI` + SIG_EST*DI;
  WI=INVERSE(VI,PSI_EST,SIG_EST);
  Q_LAMDA = Q_LAMDA + RESID_I`*WI*RESID_I;
  NCUM=NCUM+RI;
 END;
FINISH;

START SSTOTAL;     /* COMPUTES TOTAL SUMS OF SQUARES FOR SELECTED MODEL*/
 RUN MODEL;
 SS_TOTAL=0;NCUM=0;
 DO J=1 TO N;
  RI=SAMSIZ[J,];
  FIRST=NCUM+1;LAST=NCUM+RI;
  YI=Y082152[FIRST:LAST,];
  ZI=RANDOM*Z[FIRST:LAST,];
  DI=_WEIGHT_[FIRST:LAST,];
  DI=INV(DIAG(DI));
  VI = ZI*PSI_EST*ZI` + SIG_EST*DI;
  WI=INVERSE(VI,PSI_EST,SIG_EST);
  SS_TOTAL = SS_TOTAL + YI`*WI*YI;
  NCUM=NCUM+RI;
 END;
FINISH;

START PRINTER;     /* MODULE CALL FOR PRINTING SELECTED OUTCOMES       */
RESET NONAME;
 RUN CONCORR;
 RUN AIC;

NN_=(N//N_);
RRMIN=(RMIN//RMIN_);
RRMAX=(RMAX//RMAX_);
RRAVG=(RAVG//RAVG_);
TTOTAL=(TOTAL//TOTAL_);

ADJUSTN0=ADJUSTN||"PSI =";ADJUSTN=ADJUSTN||"PSI =";
MSEP="MSE =";METHODN="METHOD =";
ALPHANT=ALPHAN`;
IF ANY(OPTIONS="SHORT") THEN GOTO SKIPTHIS;
PRINT / 'MACRO PROCEDURE MIXNLIN',,
        'SAMPLE SIZE INFORMATION',,
        METHODN METHOD,,
        {"FIXED-EFFECTS","VARIANCE PARAMETERS"} (|FORMAT=$20.|)
        NN_    (|COLNAME='NUMBER OF SUBJECTS' FORMAT=9.0|)
        RRMIN  (|COLNAME='MINIMUM OBS.' FORMAT=7.0|)
        RRMAX  (|COLNAME='MAXIMUM OBS.' FORMAT=7.0|)
        RRAVG  (|COLNAME='AVERAGE OBS.' FORMAT=7.2|)
        TTOTAL (|COLNAME='TOTAL OBS.'  FORMAT=8.0|),,
       'STARTING VALUES FOR VARIANCE-COVARIANCE STRUCTURE',,
       'POOLED WITHIN-SUBJECT','ERROR MEAN SQUARE',,MSEP SIG0,,
       'ADDITIONAL VARIANCE PARAMETER',,
       'BETWEEN-SUBJECT SAMPLE','COVARIANCE MATRIX',,
        ADJUSTN0 PSI0 (|FORMAT=BEST16.|);
SKIPTHIS:
QNAME={"-2*LOG(L)"};
IF METHOD={"EGLS"} THEN QNAME={"RESIDUAL SS"};
IF METHOD={"IRGLS"} THEN QNAME={"RESIDUAL SS"};
HALFNAME={"SUBITER"};
IF METHOD={"EGLS"} THEN HALFNAME={"HALFSTEP"};
IF METHOD={"IRGLS"} THEN HALFNAME={"HALFSTEP"};
SUMNAME={'CYCLE'}||{'ITER'}||HALFNAME||ALPHANT||QNAME;
PRINT / 'MACRO PROCEDURE MIXNLIN',,
       'REPEATED MEASURES MULTIVARIATE',
       'NON-LINEAR LEAST SQUARES ITERATIVE PHASE',,
       METHODN METHOD,
       SUMMARY (|COLNAME=SUMNAME FORMAT=BEST12.|),,
       PRINT1,
       PRINT2;

STDERR=VECDIAG(OMEGA);
STDERR=SQRT(ABS(STDERR));
PARAM=ALPHA;
LCL=PARAM - TINV(.975,NROW(Y082152)-PQ)*STDERR;
UCL=PARAM + TINV(.975,NROW(Y082152)-PQ)*STDERR;
T_STAT=PARAM/STDERR;
P_VAL=2*(1-PROBT(ABS(PARAM/STDERR),NROW(Y082152)-PQ));
P_VAL=P_VAL<>.0001;
SOURCE={"REGRESSION","RESIDUAL","UNCORRECTED TOTAL"};
DF=PQ//(NROW(Y082152)-PQ)//(NROW(Y082152));
SS=(SS_TOTAL-Q_SS)//Q_SS//(SS_TOTAL);
MS=SS/DF;MS[3,1]=.;
NAMEFIT={"TOTAL OBSERVATIONS",
         "AVERAGE MODEL R-SQUARE:",
         "AVERAGE MODEL CONCORDANCE CORRELATION:",
         "CONDITIONAL MODEL R-SQUARE:",
         "CONDITIONAL MODEL CONCORDANCE CORRELATION:",
         "VARIANCE-COVARIANCE CONCORDANCE CORRELATION:",
         "APPROX. CHI-SQUARE FOR H0: COVARIANCE STRUCTURE IS CORRECT",
         "DEGREES OF FREEDOM",
         "PR > CHI-SQUARE",
         "AKAIKE'S INFORMATION CRITERION (AIC)",
         "SCHWARTZ'S BAYESIAN CRITERION (SBC)",
         "-2*LOG LIKELIHOOD"};
VALFIT = TOTAL//RSQ_PA//RHO_PA//RSQ_SS//RHO_SS//
         RHO_CVAR//LRTEST//DFLRTEST//P_LRTEST//AIC//SBC//LOGLIKE2;

       PARAM  = {" ","ESTIMATE"," "}//CHAR(PARAM,10);
       STDERR = COVAR//{"STD. ERROR"," "}//CHAR(STDERR,11);
       T_STAT = {"T FOR H0:","PARAMETER=0"," "}//CHAR(T_STAT,11,2);
       P_VAL  = {" ","PR > |T|"," "}//CHAR(P_VAL,8,4);
       LCL    = {"LOWER","95% LIMIT"," "}//CHAR(LCL,10);
       UCL    = {"UPPER","95% LIMIT"," "}//CHAR(UCL,10);

RESET SPACES=2;
PRINT / 'MACRO PROCEDURE MIXNLIN',
        METHODN METHOD,,
       'MODEL FITTING INFORMATION',
        MSG1,
        NAMEFIT (|COLNAME='DESCRIPTION' FORMAT=$60.|)
        VALFIT  (|COLNAME='VALUE' FORMAT=BEST8.|),,
        "FINAL PARAMETER ESTIMATES",,
       SOURCE (|COLNAME='SOURCE' FORMAT=$18.|)
           DF (|COLNAME='DF' FORMAT=4.0|)
           SS (|COLNAME='SUM OF SQUARES' FORMAT=BEST16.|)
           MS (|COLNAME='MEAN SQUARE' FORMAT=BEST16.|),,,
       NAMEB
       NAMEX
       VALUE
       PARAM  (|FORMAT=$10.|)
       STDERR (|FORMAT=$10.|)
       T_STAT (|FORMAT=$11.|)
       P_VAL  (|FORMAT=$8.|)
       LCL    (|FORMAT=$10.|)
       UCL    (|FORMAT=$10.|);
RESET SPACES=1;

PRINT / 'MACRO PROCEDURE MIXNLIN',,
        'FINAL PARAMETER ESTIMATES',,
        METHODN METHOD,,
       'VARIANCE-COVARIANCE STRUCTURE',,
       'POOLED WITHIN-SUBJECT','ERROR MEAN SQUARE',,MSEP SIGPOOL,,
       'VARIANCE PARAMETER',
       'BETWEEN-SUBJECT SAMPLE','COVARIANCE MATRIX',,
        ADJUSTN PSI (|FORMAT=BEST16.|);

IF ANY(OPTIONS="ALL") THEN DO;
PRINT / 'MACRO PROCEDURE MIXNLIN',,
       METHODN METHOD,,
       'ASYMPTOTIC VARIANCE-COVARIANCE MATRIX OF ESTIMATED PARAMETERS',
       OMEGA_A (|ROWNAME=ALPHAN COLNAME=ALPHANT FORMAT=BEST16.|),,
       'ROBUST VARIANCE-COVARIANCE MATRIX OF ESTIMATED PARAMETERS',
       OMEGA_R (|ROWNAME=ALPHAN COLNAME=ALPHANT FORMAT=BEST16.|);
END;
RETURN;
FINISH;

START GMANOVA;     /* GMANOVA MODULE FOR PERFORMING DEFAULT HYP. TESTS */
 ALPHA = ALPHA[LOC(VECDIAG(OMEGA)>0),];
 OMEGA = OMEGA[LOC(VECDIAG(OMEGA)>0),LOC(VECDIAG(OMEGA)>0)];

 RESULT=1:9;
 SOURCE='RESIDUAL';
NO_TESTS=1+NO_COVAR;
IF NO_CLASS=1 THEN DO;
  DFHYP=1;
  CHYP=J(1,1,1);
 IF NO_COVAR>0 THEN DO;
   DFHYP=(DFHYP)//J(NO_COVAR,1,1);
   CHYP= ( CHYP||J(1,NO_COVAR,0)      )//
         ( J(NO_COVAR,NO_CLASS,0)||I(NO_COVAR) );
  END;
END;

IF NO_CLASS>1 THEN DO;
  DFHYP=(NO_CLASS-1);
  CHYP=( J((NO_CLASS-1),1,1)||(-1*I(NO_CLASS-1)) );
 IF NO_COVAR>0 THEN DO;
   DFHYP=(DFHYP)//J(NO_COVAR,1,1);
   CHYP= ( CHYP||J((NO_CLASS-1),NO_COVAR,0)      )//
         ( J(NO_COVAR,NO_CLASS,0)||I(NO_COVAR) );
 END;
END;

NCUM=0;
DO I=1 TO NO_TESTS;
 DF=DFHYP[I,1];
 FIRST=NCUM+1;LAST=NCUM+DF;
 S0 = MIN(P-NO_RSCT,DF);
 N0 = ((N-Q)-P-NO_RSCT)/2;
 DF_ERR1=TOTAL-PQ;
 DF_ERR2=N-PQ;
 IF (N-PQ)<=0 THEN DF_ERR2=1;
 IF TYPE={"MANOVA"} THEN DF_ERR2=(2*S0*N0-S0+2);
 C=CHYP[FIRST:LAST,];
 DF_HYP=DF*(P-NO_RSCT);
 C=I((P-NO_RSCT))@C;
 T_SQ=(C*ALPHA)`*INV(C*OMEGA*C`)*(C*ALPHA);
 F_TEST1=T_SQ/(DF_HYP*(Q_SS/DF_ERR1));
 IF METHOD={"OLS"} THEN F_TEST1=T_SQ/DF_HYP;
 F_TEST2=T_SQ/DF_HYP;
 IF TYPE={"MANOVA"} THEN
    F_TEST2=(2*S0*N0-S0+2)*T_SQ/(S0*DF*(P-NO_RSCT)*(N-Q));
 P_VALUE=(1-PROBCHI(T_SQ,DF_HYP));
 P_F1=(1-PROBF(F_TEST1,DF_HYP,DF_ERR1));
 P_F2=(1-PROBF(F_TEST2,DF_HYP,DF_ERR2));
 IF P_VALUE<.0001 THEN P_VALUE=.0001;
 IF P_F1<.0001 THEN P_F1=.0001;
 IF P_F2<.0001 THEN P_F2=.0001;
 SOURCE=SOURCE//LABELS[I,];
 RESULT=
RESULT//(DF_HYP||T_SQ||P_VALUE||F_TEST1||DF_ERR1||P_F1||F_TEST2||DF_ERR2||P_F2);
 NCUM=NCUM+DF;
END;

SOURCE = SOURCE[2:NROW(SOURCE),];
RESULT = RESULT[2:NROW(RESULT),];
DF_HYP = RESULT[,1];
T_SQ=RESULT[,2];
P_VALUE=RESULT[,3];
F_TEST1=RESULT[,4];
DF_ERR1=RESULT[,5];
P_F1=RESULT[,6];
F_TEST2=RESULT[,7];
DF_ERR2=RESULT[,8];
P_F2=RESULT[,9];
FNAME={"MODIFIED F"};
IF TYPE={"MANOVA"} THEN FNAME={"MANOVA F"};
PRINT / 'MACRO PROCEDURE MIXNLIN',,
        'GENERALIZED MULTIVARIATE ANALYSIS OF VARIANCE TABLE',
        'LARGE SAMPLE TESTS OF HYPOTHESES',,
   SOURCE (|COLNAME='SOURCE'|)
   DF_HYP (|COLNAME='D.F.' FORMAT=4.0|)
   T_SQ (|COLNAME='CHI SQUARE' FORMAT=BEST10.|)
   P_VALUE (|COLNAME='P-VALUE' FORMAT=7.4|)
   F_TEST1 (|COLNAME='WALD F' FORMAT=BEST10.|)
   DF_ERR1 (|COLNAME='DF ERROR' FORMAT=8.0|)
   P_F1 (|COLNAME='P-VALUE' FORMAT=7.4|)
   F_TEST2 (|COLNAME=FNAME FORMAT=BEST10.|)
   DF_ERR2 (|COLNAME='DF ERROR' FORMAT=8.0|)
   P_F2 (|COLNAME='P-VALUE' FORMAT=7.4|);
FINISH;

START GMANOVAI;    /* GMANOVA MODULE FOR PERFORMING DEFAULT HYP. TESTS */
                   /* FOR EACH SELECTED PARAMETER.                     */
 ALPHA = ALPHA[LOC(VECDIAG(OMEGA)>0),];
 OMEGA = OMEGA[LOC(VECDIAG(OMEGA)>0),LOC(VECDIAG(OMEGA)>0)];
 RESULT=1:9;
 SOURCE='RESIDUAL';
NO_TESTS=1+NO_COVAR;

IF NO_CLASS=1 THEN DO;
  DFHYP=1;
  CHYP=J(1,1,1);
 IF NO_COVAR>0 THEN DO;
   DFHYP=(DFHYP)//J(NO_COVAR,1,1);
   CHYP= ( CHYP||J(1,NO_COVAR,0)      )//
         ( J(NO_COVAR,NO_CLASS,0)||I(NO_COVAR) );
  END;
END;

IF NO_CLASS>1 THEN DO;
  DFHYP=(NO_CLASS-1);
  CHYP=( J((NO_CLASS-1),1,1)||(-1*I(NO_CLASS-1)) );
 IF NO_COVAR>0 THEN DO;
   DFHYP=(DFHYP)//J(NO_COVAR,1,1);
   CHYP= ( CHYP||J((NO_CLASS-1),NO_COVAR,0)      )//
         ( J(NO_COVAR,NO_CLASS,0)||I(NO_COVAR) );
 END;
END;

A=I((P-NO_RSCT));
NCUM=0;
NAMETEST=REPEAT(NAMETEST,NO_TESTS,1);
DO I=1 TO NO_TESTS;
 DF=DFHYP[I,1];
 FIRST=NCUM+1;LAST=NCUM+DF;
DO SUB=1 TO (P-NO_RSCT);
 S0 = MIN(1,DF);
 N0 = ((N-Q)-1)/2;
 DF_ERR1=TOTAL-PQ;
 DF_ERR2=N-PQ;
 IF (N-PQ)<=0 THEN DF_ERR2=1;
 IF TYPE={"MANOVA"} THEN DF_ERR2=(2*S0*N0-S0+2);
 CTEST=CHYP[FIRST:LAST,];
 DF_HYP=DF*1;
 C=A[SUB,]@CTEST;
 T_SQ=(C*ALPHA)`*INV(C*OMEGA*C`)*(C*ALPHA);
 F_TEST1=T_SQ/(DF_HYP*(Q_SS/DF_ERR1));
 IF METHOD={"OLS"} THEN F_TEST1=T_SQ/DF_HYP;
 F_TEST2=T_SQ/DF_HYP;
 IF TYPE={"MANOVA"} THEN
    F_TEST2=(2*S0*N0-S0+2)*T_SQ/(S0*DF*1*(N-Q));
 P_VALUE=(1-PROBCHI(T_SQ,DF_HYP));
 P_F1=(1-PROBF(F_TEST1,DF_HYP,DF_ERR1));
 P_F2=(1-PROBF(F_TEST2,DF_HYP,DF_ERR2));
 IF P_VALUE<.0001 THEN P_VALUE=.0001;
 IF P_F1<.0001 THEN P_F1=.0001;
 IF P_F2<.0001 THEN P_F2=.0001;
 SOURCE=SOURCE//LABELS[I,];
 RESULT=
RESULT//(DF_HYP||T_SQ||P_VALUE||F_TEST1||DF_ERR1||P_F1||F_TEST2||DF_ERR2||P_F2);
END;
 NCUM=NCUM+DF;
END;
SOURCE = SOURCE[2:NROW(SOURCE),];
RESULT = RESULT[2:NROW(RESULT),];
DF_HYP = RESULT[,1];
T_SQ=RESULT[,2];
P_VALUE=RESULT[,3];
F_TEST1=RESULT[,4];
DF_ERR1=RESULT[,5];
P_F1=RESULT[,6];
F_TEST2=RESULT[,7];
DF_ERR2=RESULT[,8];
P_F2=RESULT[,9];
FNAME={"MODIFIED F"};
IF TYPE={"MANOVA"} THEN FNAME={"MANOVA F"};
PRINT / 'MACRO PROCEDURE MIXNLIN',,
        'GENERALIZED MULTIVARIATE ANALYSIS OF VARIANCE TABLE',
        'LARGE SAMPLE TESTS OF HYPOTHESES',,
   SOURCE (|COLNAME='SOURCE'|)
   NAMETEST (|COLNAME='PARAMETER'|)
   DF_HYP (|COLNAME='D.F.' FORMAT=4.0|)
   T_SQ (|COLNAME='CHI SQUARE' FORMAT=BEST10.|)
   P_VALUE (|COLNAME='P-VALUE' FORMAT=7.4|)
   F_TEST1 (|COLNAME='WALD F' FORMAT=BEST10.|)
   DF_ERR1 (|COLNAME='DF ERROR' FORMAT=8.0|)
   P_F1 (|COLNAME='P-VALUE' FORMAT=7.4|)
   F_TEST2 (|COLNAME=FNAME FORMAT=BEST10.|)
   DF_ERR2 (|COLNAME='DF ERROR' FORMAT=8.0|)
   P_F2 (|COLNAME='P-VALUE' FORMAT=7.4|);
FINISH;

START CONTRAST;         /* CONTRAST MODULE FOR PERFORMING USER DEFINED TESTS */
IF NO_RSCT>0 THEN GOTO ENDCONT;
 RESULT=1:9;
 SOURCE=CSOURCE[2:NO_C,];
 CHYP=CONTRAST[2:NROW(CONTRAST),];
 DFHYP=RANK_C[2:NO_C,];
 NO_TESTS=NO_C-1;
NCUM=0;
DO I=1 TO NO_TESTS;
 DF=DFHYP[I,1];
 FIRST=NCUM+1;LAST=NCUM+DF;
 S0 = MIN(1,DF);
 N0 = ((N-Q)-1)/2;
 DF_ERR1=TOTAL-PQ;
 DF_ERR2=N-PQ;
 IF (N-PQ)<=0 THEN DF_ERR2=1;
*IF TYPE={"MANOVA"} THEN DF_ERR2=(2*S0*N0-S0+2); /* modify at a later date */
 C=CHYP[FIRST:LAST,];
 DF_HYP=DF;
 T_SQ=(C*ALPHA)`*INV(C*OMEGA*C`)*(C*ALPHA);
 F_TEST1=T_SQ/(DF_HYP*(Q_SS/DF_ERR1));
 IF METHOD={"OLS"} THEN F_TEST1=T_SQ/DF_HYP;
 F_TEST2=T_SQ/DF_HYP;
*IF TYPE={"MANOVA"} THEN
    F_TEST2=(2*S0*N0-S0+2)*T_SQ/(S0*DF*P*(N-Q)); /* modify at a later date */
 P_VALUE=(1-PROBCHI(T_SQ,DF_HYP));
 P_F1=(1-PROBF(F_TEST1,DF_HYP,DF_ERR1));
 P_F2=(1-PROBF(F_TEST2,DF_HYP,DF_ERR2));
 IF P_VALUE<.0001 THEN P_VALUE=.0001;
 IF P_F1<.0001 THEN P_F1=.0001;
 IF P_F2<.0001 THEN P_F2=.0001;
 RESULT=
RESULT//(DF_HYP||T_SQ||P_VALUE||F_TEST1||DF_ERR1||P_F1||F_TEST2||DF_ERR2||P_F2);
 NCUM=NCUM+DF;
END;

RESULT = RESULT[2:NROW(RESULT),];
DF_HYP = RESULT[,1];
T_SQ=RESULT[,2];
P_VALUE=RESULT[,3];
F_TEST1=RESULT[,4];
DF_ERR1=RESULT[,5];
P_F1=RESULT[,6];
F_TEST2=RESULT[,7];
DF_ERR2=RESULT[,8];
P_F2=RESULT[,9];
FNAME={"MODIFIED F"};
*IF TYPE={"MANOVA"} THEN FNAME={"MANOVA F"}; /* modify at a later date */
SOURCE=TRIM(SUBSTR(SOURCE,1,40));
PRINT / 'MACRO PROCEDURE MIXNLIN',,
        'GENERALIZED MULTIVARIATE ANALYSIS OF VARIANCE TABLE',
        'LARGE SAMPLE TESTS OF USER SPECIFIED HYPOTHESES',,
   SOURCE (|COLNAME='SOURCE'|)
   DF_HYP (|COLNAME='D.F.' FORMAT=4.0|)
   T_SQ (|COLNAME='CHI SQUARE' FORMAT=BEST10.|)
   P_VALUE (|COLNAME='P-VALUE' FORMAT=7.4|)
   F_TEST1 (|COLNAME='WALD F' FORMAT=BEST10.|)
   DF_ERR1 (|COLNAME='DF ERROR' FORMAT=8.0|)
   P_F1 (|COLNAME='P-VALUE' FORMAT=7.4|)
   F_TEST2 (|COLNAME=FNAME FORMAT=BEST10.|)
   DF_ERR2 (|COLNAME='DF ERROR' FORMAT=8.0|)
   P_F2 (|COLNAME='P-VALUE' FORMAT=7.4|);
ENDCONT:
 IF NO_RSCT>0 THEN DO;
 PRINT "RESTRICT Statement in effect: Test not done";
 END;
FINISH;

START FUNCTION; /* SUBROUTINE FOR DEFINING ESTIMABLE FUNCTIONS OF ALPHA */
IF NO_RSCT>0 THEN GOTO ENDFUNC;
 RESULT=1:7;
 SOURCE=ESOURCE[2:NO_E,];
 EHYP=ESTIMAT[2:NROW(ESTIMAT),];
 DFHYP=RANK_E[2:NO_E,];
 NO_EST=NO_E-1;
NCUM=0;
DO I=1 TO NO_EST;
 DF=DFHYP[I,1];
 FIRST=NCUM+1;LAST=NCUM+DF;
 S0 = MIN(1,DF);
 N0 = ((N-Q)-1)/2;
 DF_ERR1=TOTAL-PQ;
 DF_ERR2=N-PQ;
 IF (N-PQ)<=0 THEN DF_ERR2=1;
*IF TYPE={"MANOVA"} THEN DF_ERR2=(2*S0*N0-S0+2); /* modify at a later date */
 C=EHYP[FIRST:LAST,];
 DF_HYP=DF;
 EST=C*ALPHA;
 SE = SQRT(C*OMEGA*C`);
 T_STAT=EST/SE;
 P_VAL1=2*(1-PROBT(ABS(EST/SE),DF_ERR1));
 P_VAL1=P_VAL1<>.0001;
 P_VAL2=2*(1-PROBT(ABS(EST/SE),DF_ERR2));
 P_VAL2=P_VAL2<>.0001;
 RESULT=
RESULT//(EST||SE||T_STAT||DF_ERR1||P_VAL1||DF_ERR2||P_VAL2);
 NCUM=NCUM+DF;
END;

FNAME={"MODIFIED DF"};
*IF TYPE={"MANOVA"} THEN FNAME={"MANOVA DF"}; /* modify at a later date */

RESULT = RESULT[2:NROW(RESULT),];
EST = RESULT[,1];
SE  =RESULT[,2];
T_STAT=RESULT[,3];
DF_ERR1=RESULT[,4];
P_VAL1=RESULT[,5];
DF_ERR2=RESULT[,6];
P_VAL2=RESULT[,7];
SOURCE=TRIM(SUBSTR(SOURCE,1,40));
PRINT / 'MACRO PROCEDURE MIXNLIN',,
        'SUMMARY STATISTICS ON USER SPECIFIED ESTIMABLE FUNCTIONS',,
   SOURCE (|COLNAME='EFFECT'|)
   EST (|COLNAME='ESTIMATE' FORMAT=BEST10.|)
   SE  (|COLNAME='STD. ERR.' FORMAT=BEST10.|)
   T_STAT (|COLNAME='T: EFFECT=0' FORMAT=BEST10.|)
   DF_ERR1 (|COLNAME='WALD DF' FORMAT=8.0|)
   P_VAL1 (|COLNAME='P-VALUE' FORMAT=7.4|)
   DF_ERR2 (|COLNAME=FNAME FORMAT=11.0|)
   P_VAL2 (|COLNAME='P-VALUE' FORMAT=7.4|);
 ENDFUNC:
 IF NO_RSCT>0 THEN DO;
 PRINT "RESTRICT Statement in effect: Estimation not done";
 END;
FINISH;


START OUTPUT;      /* OUTPUT  MODULE FOR SENDING OUTPUT TO A DATA SET  */

B082152 = REPEAT(0,TOTAL,P); * INITIALIZE B082152=(TOTAL X P)MATRIX OF ZEROES
                               FOR EVALUATING X[I,], Z[I,], AND DI AT THE
                               FIXED POPULATION AVERAGED VALUE OF BETA_I;
RUN MODEL;
RUN RANDOM;
VAR_MEAN=J(TOTAL,1,0);
VAR_PRED=J(TOTAL,1,0);
DO I=1 TO TOTAL;
VAR_MEAN[I,] = X[I,]*_OMEGA_*X[I,]`;
VAR_PRED[I,]=Z[I,]*PSI*Z[I,]` + SIGPOOL/_WEIGHT_[I,] + VAR_MEAN[I,];
END;
  ZB_HAT=J(TOTAL,1,0);
 DO JJ=1 TO TOTAL;
  ZB_HAT[JJ,]=Z[JJ,]*BI_HAT[JJ,]`;
 END;
 BI_NAME=1:V;
 BI_NAME=TRIM(CHAR(BI_NAME));
 BI_NAME=CONCAT( J(1,V,"BHAT_"),BI_NAME );
 BI_NAME=COMPRESS(BI_NAME);
 BIHAT=BI_HAT;
 IF V<12 THEN BIHAT=BIHAT||J(TOTAL,(12-V),0);
 BHAT_1=BIHAT[,1];
 BHAT_2=BIHAT[,2];
 BHAT_3=BIHAT[,3];
 BHAT_4=BIHAT[,4];
 BHAT_5=BIHAT[,5];
 BHAT_6=BIHAT[,6];
 BHAT_7=BIHAT[,7];
 BHAT_8=BIHAT[,8];
 BHAT_9=BIHAT[,9];
 BHAT_10=BIHAT[,10];
 BHAT_11=BIHAT[,11];
 BHAT_12=BIHAT[,12];

STUDENT=TINV((1-TYPE_I/2),(TOTAL-PQ));
CLM_LOW=_PRED_ - STUDENT*SQRT(VAR_MEAN);
CLM_UP =_PRED_ + STUDENT*SQRT(VAR_MEAN);
CLI_LOW=_PRED_ - STUDENT*SQRT(VAR_PRED);
CLI_UP =_PRED_ + STUDENT*SQRT(VAR_PRED);

EDIT R082152;
SETOUT R082152;
REPLACE ALL VAR {"_PRED_" "CLM_LOW" "CLM_UP" "CLI_LOW" "CLI_UP" "PRED_SS"};
REPLACE ALL VAR BI_NAME;
FINISH;


START NLIN;        /* NLIN REGRESSION MODULE USED FOR OLS, EGLS, IRGLS */

/* INITIALIZE ALPHA0, PSI AND SIGMA2 TO BE THE EGLS ESTIMATORS */
 ALPH_EST=ALPHA0  ;
 PSI_EST=PSI0    ;
 SIG_EST=SIG0    ;
/*                  END OF INITIALIZATION                    */

/* PHASE 1: INITIAL OLS ESTIMATION */
/* PHASE 2: PROCEED IN USUAL FASHION */
NCYCLES=1 + RANDOM*(NCYCLE-1);
RUN MODEL;
RUN RANDOM;
RUN ESTIMATE; /* THIS RUN OF ESTIMATE IS ONLY TO COMPUTE Q_SS FOR CYCLE=0 */
CYCLE=0;
ITER=0;
STEP=0;
SUMMARY=CYCLE||ITER||STEP||ALPH_EST`||Q_SS;
DO CYCLE=1 TO NCYCLES;
 SIG_TEMP= (1-SIGSQIND)*SIG_EST + SIGSQIND*SIGSQ;
 IF (CYCLE>1)&(RANDOM=1)=1 THEN DO;
  RUN MOMENTS; /* FOR CURRENT ESTIMATES OF (ALPHA), UPDATES (PSI SIG) */
  PSI_EST = PSI_GLS;
  SIG_EST = SIG;
 END;

 ADJUST=0;
 EPISOLON=CONVERGE;
 TAU=10##-3;
 Q_SS0=0;
 Q_LAMDA=0;
 RUN ESTIMATE;  /* UPDATES F, X, DELTA | CURRENT (ALPHA BI_GLS PSI SIG) */
 RUN ESTIM_BI;  /* UPDATES BI_GLS | CURRENT (ALPHA BI_GLS PSI SIG) */
 ITER=0;
 STEP=0;
PRINT1="NOTE: CONVERGENCE CRITERION MET.";
PRINT2=" ";

DO ITER=1 TO NLITER BY 1 WHILE(ABS((Q_SS-Q_SS0)/(Q_SS0+TAU))>EPISOLON);
 Q_SS0 = Q_SS;
 Q_LAMDA=Q_SS;
 ALPHA_L=ALPH_EST;

IF STEPS={"NOHALVE"} THEN DO;
 ALPH_EST = ALPHA_L + DELTA;
END;

ELSE IF (METHOD={"EGLS"})&(RANDOM=1)&(S082152=1)&(ITER=1)=1 THEN DO;
 ALPH_EST= ALPHA_L + DELTA;
END;

ELSE DO;
DO STEP=1 TO 10 BY 1 WHILE(Q_LAMDA>=Q_SS);
 LAMDA=((1/2)**(STEP-1));
 ALPH_EST = ALPHA_L + LAMDA*DELTA;
RUN SEARCH;
IF STEP=10 THEN DO;
ALPH_EST=ALPHA_L;
PRINT1="WARNING: STEP HALVING PERFORMED 10 TIMES WITH NO IMPROVEMENT.";
PRINT2="         CONVERGENCE ASSUMED.                                ";
END;
END;
END;

IF RANDOM=0 THEN DO;
SIG_TEMP= (1-SIGSQIND)*Q_SS/(TOTAL-PQ) + SIGSQIND*SIGSQ;
END;

RUN ESTIMATE;
STEP=STEP-1;
SUMMARY=SUMMARY//(CYCLE||ITER||STEP||ALPH_EST`||Q_SS);
  IF ITER=NLITER THEN DO;
   PRINT1="WARNING: THE MAXIMUM NUMBER OF ITERATIONS ALLOWED HAVE BEEN  ";
   PRINT2="         REACHED. CONVERGENCE IS ASSUMED.                    ";
  END;

END;

END; * THIS END STATEMENT IS FOR CYCLE;

RUN SSTOTAL;

IF RANDOM=0 THEN GOTO OLS;
IF RANDOM=1 THEN GOTO EGLS;

OLS:
 ADJUSTN=NAME(UNADJUSTED);
 ALPHANT=ALPHAN`;
 ALPH_OLS=ALPH_EST;
 SIG_OLS=(1-SIGSQIND)*Q_SS/(TOTAL-PQ) + SIGSQIND*SIGSQ;
 PSI_OLS=J(V0,V0,0);
 VAR_OLS=SIG_OLS*OMEGA;
 ALPHA=ALPH_OLS;
 PSI=PSI_OLS;
 SIGPOOL=SIG_OLS;
 OMEGA=VAR_OLS;
 RUN ROBUST;
IF COVAR={"ROBUST"} THEN OMEGA=OMEGA_R;
RESET NONAME;

IF METHOD={"OLS"} THEN GOTO PRINTOLS;
IF ANY(OPTIONS="SHORT") THEN GOTO SKIPOLS;

PRINTOLS:
RUN CONCORR;
SUMNAME={'CYCLE' 'ITER' 'HALFSTEP'}||ALPHANT||{'RESIDUAL SS'};
PRINT / 'MACRO PROCEDURE MIXNLIN',,
       'REPEATED MEASURES MULTIVARIATE',
      'NON-LINEAR LEAST SQUARES ITERATIVE PHASE',,
      'STAGE 1 ESTIMATION: OLS',,
       SUMMARY (|COLNAME=SUMNAME FORMAT=BEST12.|),,
       PRINT1,
       PRINT2;

STDERR=VECDIAG(OMEGA);
STDERR=SQRT(ABS(STDERR));
PARAM=ALPHA;
LCL=PARAM - TINV(.975,NROW(Y082152)-PQ)*STDERR;
UCL=PARAM + TINV(.975,NROW(Y082152)-PQ)*STDERR;
T_STAT=PARAM/STDERR;
P_VAL=2*(1-PROBT(ABS(PARAM/STDERR),NROW(Y082152)-PQ));
P_VAL=P_VAL<>.0001;
SOURCE={"REGRESSION","RESIDUAL","UNCORRECTED TOTAL"};
DF=PQ//(NROW(Y082152)-PQ)//(NROW(Y082152));
SS=(SS_TOTAL-Q_SS)//Q_SS//(SS_TOTAL);
MS=SS/DF;MS[3,1]=.;
RUN AIC;
NAMEFIT={"TOTAL OBSERVATIONS",
         "AVERAGE MODEL R-SQUARE:",
         "AVERAGE MODEL CONCORDANCE CORRELATION:",
         "CONDITIONAL MODEL R-SQUARE:",
         "CONDITIONAL MODEL CONCORDANCE CORRELATION:",
         "VARIANCE-COVARIANCE CONCORDANCE CORRELATION:",
         "APPROX. CHI-SQUARE FOR H0: COVARIANCE STRUCTURE IS CORRECT",
         "DEGREES OF FREEDOM",
         "PR > CHI-SQUARE",
         "AKAIKE'S INFORMATION CRITERION (AIC)",
         "SCHWARTZ'S BAYESIAN CRITERION (SBC)",
         "-2*LOG LIKELIHOOD"};
VALFIT = TOTAL//RSQ_PA//RHO_PA//RSQ_SS//RHO_SS//
         RHO_CVAR//LRTEST//DFLRTEST//P_LRTEST//AIC//SBC//LOGLIKE2;

       PARAM  = {" ","ESTIMATE"," "}//CHAR(PARAM,10);
       STDERR = COVAR//{"STD. ERROR"," "}//CHAR(STDERR,10);
       T_STAT = {"T FOR H0:","PARAMETER=0"," "}//CHAR(T_STAT,11,2);
       P_VAL  = {" ","PR > |T|"," "}//CHAR(P_VAL,8,4);
       LCL    = {"LOWER","95% LIMIT"," "}//CHAR(LCL,10);
       UCL    = {"UPPER","95% LIMIT"," "}//CHAR(UCL,10);

RESET SPACES=2;
PRINT / 'MACRO PROCEDURE MIXNLIN',,
        'MODEL FITTING INFORMATION',
        MSG1,
        NAMEFIT (|COLNAME='DESCRIPTION' FORMAT=$60.|)
        VALFIT  (|COLNAME='VALUE'  FORMAT=BEST8.|),,
        "FINAL PARAMETER ESTIMATES",,
       'STAGE 1 ESTIMATION: OLS',,
       SOURCE (|COLNAME='SOURCE' FORMAT=$18.|)
           DF (|COLNAME='DF' FORMAT=4.0|)
           SS (|COLNAME='SUM OF SQUARES' FORMAT=BEST16.|)
           MS (|COLNAME='MEAN SQUARE' FORMAT=BEST16.|),,,
       NAMEB
       NAMEX
       VALUE
       PARAM  (|FORMAT=$10.|)
       STDERR (|FORMAT=$10.|)
       T_STAT (|FORMAT=$11.|)
       P_VAL  (|FORMAT=$8.|)
       LCL    (|FORMAT=$10.|)
       UCL    (|FORMAT=$10.|);
RESET SPACES=1;

IF METHOD={"OLS"} THEN IF ANY(OPTIONS="ALL") THEN DO;
PRINT / 'MACRO PROCEDURE MIXNLIN',,
       METHODN METHOD,,
       'ASYMPTOTIC VARIANCE-COVARIANCE MATRIX OF ESTIMATED PARAMETERS',
       OMEGA_A (|ROWNAME=ALPHAN COLNAME=ALPHANT FORMAT=BEST16.|),,
       'ROBUST VARIANCE-COVARIANCE MATRIX OF ESTIMATED PARAMETERS',
       OMEGA_R (|ROWNAME=ALPHAN COLNAME=ALPHANT FORMAT=BEST16.|);
END;
RETURN;

SKIPOLS:
RETURN;
RETURN;

EGLS:
 ADJUST=ADJUSTGL;
 ALPH_GLS=ALPH_EST;
 SIG_GLS=SIG_EST;
 PSI_GLS=PSI_EST;
 VAR_GLS=OMEGA;
ALPHA=ALPH_GLS;
PSI=PSI_GLS;
SIGPOOL=SIG_GLS;
OMEGA=VAR_GLS;
RUN ROBUST;
IF COVAR={"ROBUST"} THEN OMEGA=OMEGA_R;

IF METHOD={"EGLS"} THEN DO;

RUN PRINTER;

IF WARNING1=0 THEN DO;
 IF WARNING=0 THEN DO;
  RUN GMANOVA;
  RUN GMANOVAI;
 END;
 ELSE DO;
 PRINT / 'WARNING: THE GMANOVA PROCEDURE COULD NOT BE EXECUTED. THE    ',
         '         CLASS VARIABLE MUST BE SPECIFIED AS THE FIRST       ',
         '         INDEPENDENT VARIABLE LISTED IN THE DESIGN STATEMENT.';
 END;
END;

ELSE DO;
 PRINT / "WARNING: THE OVERALL DESIGN MATRIX SPECIFIED BY THE ALL = STATEMENT",
         "         IS NOT EQUIVALENT ACROSS EACH PARAMETER, B1,B2,...  THERE-",
         "         FORE, ONLY USER-SPECIFIED TESTS OF HYPOTHESES USING THE   ",
         "         TEST STATEMENTS WILL BE PERFORMED.                        ";
END;

IF NO_E>1 THEN DO;
 RUN FUNCTION;
END;
IF NO_C>1 THEN DO;
 RUN CONTRAST;
END;

IF ANY(OPTIONS="OUT") THEN DO;
 RUN OUTPUT;
END;

END;

RETURN;

FINISH;

START IRGLS;       /* IRGLS MODULE FOR COMPUTING IRGLS ESTIMATES       */

/* INITIALIZE ALPHA0, PSI AND SIGMA2 TO BE THE EGLS ESTIMATORS */
RUN NLIN;
SUMMARY=SUMMARY[1,]//SUMMARY[NROW(SUMMARY),];
 ALPH_EST=ALPH_GLS;
RUN MOMENTS;
 PSI_EST=PSI_GLS;
 SIG_EST=SIG;
/*                    END OF INITIALIZATION                    */
RUN MODEL;
RUN RANDOM;
 ADJUST=0;
 EPISOLON=CONVERGE;
 TAU=10##-3;
 Q_SS0=0;
 Q_LAMDA=0;
 RUN ESTIMATE;
 CYCLE=1;
 ITER=0;
 STEP=0;
SUMMARY=CYCLE||ITER||STEP||ALPH_EST`||Q_SS;
PRINT1="NOTE: CONVERGENCE CRITERION MET.";
PRINT2=" ";

DO ITER=1 TO MAXITER BY 1 WHILE(ABS((Q_SS-Q_SS0)/(Q_SS0+TAU))>EPISOLON);
 CYCLE=2;
 Q_SS0 = Q_SS;
 Q_LAMDA=Q_SS;
 ALPHA_L=ALPH_EST;

RUN MOMENTS;
SIG_EST=SIG;
PSI_EST=PSI_GLS;
ALPH_EST=ALPHA_L + DELTA;
RUN MODEL;
RUN RANDOM;
SIG_TEMP=SIG_EST;
RUN ESTIMATE;
RUN ESTIM_BI;
STEP=STEP-1;
SUMMARY=SUMMARY//(CYCLE||ITER||STEP||ALPH_EST`||Q_SS);
  IF ITER=MAXITER THEN DO;
   PRINT1="WARNING: THE MAXIMUM NUMBER OF ITERATIONS ALLOWED HAVE BEEN  ";
   PRINT2="         REACHED. CONVERGENCE IS ASSUMED.                    ";
  END;
END;
RUN SSTOTAL;

IRGLS:
 ADJUST=ADJUSTGL;
 ALPH_ILS=ALPH_EST;
 SIG_ILS=SIG_EST;
 PSI_ILS=PSI_EST;
 VAR_ILS=OMEGA;
ALPHA=ALPH_ILS;
PSI=PSI_ILS;
SIGPOOL=SIG_ILS;
OMEGA=VAR_ILS;
RUN ROBUST;
IF COVAR={"ROBUST"} THEN OMEGA=OMEGA_R;
RUN PRINTER;

IF WARNING1=0 THEN DO;
 IF WARNING=0 THEN DO;
  RUN GMANOVA;
  RUN GMANOVAI;
 END;
 ELSE DO;
 PRINT / 'WARNING: THE GMANOVA PROCEDURE COULD NOT BE EXECUTED. THE    ',
         '         CLASS VARIABLE MUST BE SPECIFIED AS THE FIRST       ',
         '         INDEPENDENT VARIABLE LISTED IN THE DESIGN STATEMENT.';
 END;
END;

ELSE DO;
 PRINT / "WARNING: THE OVERALL DESIGN MATRIX SPECIFIED BY THE ALL = STATEMENT",
         "         IS NOT EQUIVALENT ACROSS EACH PARAMETER, B1,B2,...  THERE-",
         "         FORE, ONLY USER-SPECIFIED TESTS OF HYPOTHESES USING THE   ",
         "         TEST STATEMENTS WILL BE PERFORMED.                        ";
END;

IF NO_E>1 THEN DO;
 RUN FUNCTION;
END;
IF NO_C>1 THEN DO;
 RUN CONTRAST;
END;

IF ANY(OPTIONS="OUT") THEN DO;
 RUN OUTPUT;
END;
RETURN;
FINISH;

START MOMENTS;     /* MOMENTS MODULE FOR CALCULATING METHOD OF MOMENTS */
                   /* ESTIMATES OF THE VARIANCE-COVARIANCE COMPONENTS  */

 DENOM=(N-Q);
 RUN MODEL;
 RUN RANDOM;
  RESID=Y082152-F;
  BB082152 = B082152;
  B=J(N,V0,0);

** FOR ESTIMATION WHEN RI<=V0 **;
  BADN=0;BADRI=0;
  SIG=0;INV_ZPZ=0;NCUM=0;
  N_    =0;
  RMIN_ =100000000;
  RMAX_ =0;
  RAVG_ =0;
  TOTAL_=0;

 DO J=1 TO N;
  RI=SAMSIZ[J,];

** FOR ESTIMATION WHEN RI<=V0 **;
 IF RI<=V0-SIG82152 THEN BADN=BADN+1;
 IF RI<=V0-SIG82152 THEN BADRI=BADRI+RI;
 IF RI<=V0-SIG82152 THEN GOTO SKIPV0;
 RMIN_=MIN(RMIN_,RI);
 RMAX_=MAX(RMAX_,RI);

  FIRST=NCUM+1;LAST=NCUM+RI;

  ZI=Z[FIRST:LAST,];
  DI=_WEIGHT_[FIRST:LAST,];
  DI=DIAG(DI);
  INV_ZPZ = INV_ZPZ +
            (1-X82152[J,]*INV(X82152`*X82152)*X82152[J,]`)*INV(ZI`*DI*ZI);
  RESID_I=RESID[FIRST:LAST,];

  BETAI_I=B082152[FIRST,]*SELECT`*S082152;
  RESID_I=RESID_I+ZI*BETAI_I`;
  BI=INV(ZI`*DI*ZI)*ZI`*DI*RESID_I;
  SI=RESID_I`*(DI-DI*ZI*INV(ZI`*DI*ZI)*ZI`*DI)*RESID_I;
  B[J,]= BI`;
  SIG = SIG + SI;
 SKIPV0:
  NCUM=NCUM+RI;
 END;
  N_ = N-BADN;
  TOTAL_ = TOTAL-BADRI;
  RAVG_ = TOTAL_/N_;
  DENOM=DENOM-BADN; ** FOR BAD RI **;
  SIG = SIG/(TOTAL-BADRI-(N-BADN)*V0+SIG82152); ** FOR BAD RI **;
  SIG=(1-SIGSQIND)*SIG + SIGSQIND*SIGSQ; ***NEWEST;
  SB = B`*( I(N) - X82152*INV(X82152`*X82152)*X82152` )*B/DENOM;
  SB = (SB+SB`)/2;
  SB_MLE=DENOM*SB/(N-BADN);
  SB_MLE=(SB_MLE+SB_MLE`)/2;
  ZROOT=INV_ZPZ/DENOM;
 CALL GENEIG(LAMDA,EIGVECL,SB,ZROOT);
  LAMDA=MIN(LAMDA);
  SIG_ADJ=MIN((SIG||LAMDA));
 CALL GENEIG(LAMDAM,EIGVECM,SB_MLE,ZROOT);
  LAMDAM=MIN(LAMDAM);
  SIG_ADJ=MIN((SIG||LAMDA));
  SIG_ADJM=MIN((SIG||LAMDAM));
  IF LAMDA<SIG THEN ADJUSTGL=1;
  ELSE ADJUSTGL=0;
  IF LAMDA>SIG THEN ADJUSTN=NAME(UNADJUSTED);
  ELSE ADJUSTN=NAME(ADJUSTED);
  IF LAMDAM<SIG THEN ADJUSTM=1;
  ELSE ADJUSTM=0;
  PSI_GLS = SB - SIG_ADJ*INV_ZPZ/DENOM;
  PSI_MLE = SB_MLE - SIG_ADJM*INV_ZPZ/DENOM;
BETATEMP=B;
FINISH;

START MLE;         /* MLE MODULE FOR FINDING MLE ESTIMATES USING THE MLE */
                   /* E-M ALGORITHM NESTED WITHIN A LEAST SQUARES LOOP   */

  N_    =N;
  RMIN_ =RMIN;
  RMAX_ =RMAX;
  RAVG_ =RAVG;
  TOTAL_=TOTAL;

/* INITIALIZE ALPHA0, PSI AND SIGMA2 TO BE THE EGLS ESTIMATORS */
 MLE_IND=1;
 ALPH_MLE=ALPH_OLS;
 PSI_MLE=PSI_MLE  ;
 SIG_MLE=SIG      ;
/*                  END OF INITIALIZATION                    */
 ADJUST=ADJUSTM;
 M2LNL=1000;M2LNL1=0;
CYCLE=1;
ITER=0;
STEP=0;
SUMMARY=CYCLE||ITER||STEP||ALPH_MLE`||M2LNL;
*PRINT M2LNL ALPH_MLE PSI_MLE SIG_MLE;
PRINT1="NOTE: CONVERGENCE CRITERION MET.";
PRINT2=" ";
DO ITER=1 TO MAXITER WHILE(ABS(M2LNL-M2LNL1)>CONVERGE);
 CYCLE=2;
 ALPH_EST=ALPH_MLE;
 M2LNL1=M2LNL;
 RUN EM_MLE;
 ALPH_MLE=ALPH_MLE+ALPH;
 STEP=SUBITERN-1;
SUMMARY=SUMMARY//(CYCLE||ITER||STEP||ALPH_MLE`||M2LNL);
  IF ITER=MAXITER THEN DO;
   PRINT1="WARNING: THE MAXIMUM NUMBER OF ITERATIONS ALLOWED HAVE BEEN  ";
   PRINT2="         REACHED. CONVERGENCE IS ASSUMED.                    ";
  END;
END;
ALPH_EST=ALPH_MLE;
PSI_EST=PSI_MLE;
SIG_EST=SIG_MLE;
RUN SSTOTAL;
ALPHA=ALPH_MLE;
PSI=PSI_MLE;
SIGPOOL=SIG_MLE;
OMEGA=VAR_MLE;
RUN ROBUST;
IF COVAR={"ROBUST"} THEN OMEGA=OMEGA_R;
RUN PRINTER;
IF WARNING1=0 THEN DO;
 IF WARNING=0 THEN DO;
  RUN GMANOVA;
  RUN GMANOVAI;
 END;
 ELSE DO;
 PRINT / 'WARNING: THE GMANOVA PROCEDURE COULD NOT BE EXECUTED. THE    ',
         '         CLASS VARIABLE MUST BE SPECIFIED AS THE FIRST       ',
         '         INDEPENDENT VARIABLE LISTED IN THE DESIGN STATEMENT.';
 END;
END;

ELSE DO;
 PRINT / "WARNING: THE OVERALL DESIGN MATRIX SPECIFIED BY THE ALL = STATEMENT",
         "         IS NOT EQUIVALENT ACROSS EACH PARAMETER, B1,B2,...  THERE-",
         "         FORE, ONLY USER-SPECIFIED TESTS OF HYPOTHESES USING THE   ",
         "         TEST STATEMENTS WILL BE PERFORMED.                        ";
END;

IF NO_E>1 THEN DO;
 RUN FUNCTION;
END;
IF NO_C>1 THEN DO;
 RUN CONTRAST;
END;

IF ANY(OPTIONS="OUT") THEN DO;
 RUN OUTPUT;
END;
FINISH;

START EM_MLE;      /* THE MLE E-M ALGORITHM FOR LINEAR MIXED-EFFECTS MODELS */

 RUN MODEL;
 RUN RANDOM;
 SIG_TEMP=SIG_MLE;
 RESID=Y082152-F;
 BB082152 = B082152;
   M2LNL=1000;M2LNL0=0;
DO SUBITERN=1 TO SUBITER WHILE(ABS(M2LNL-M2LNL0)>CONVERGE);
  XWX=0;XWY=0;YWY=0;LNDETERM=0;NCUM=0;
 DO J=1 TO N;
  RI=SAMSIZ[J,];
  FIRST=NCUM+1;LAST=NCUM+RI;
  RESID_I=RESID[FIRST:LAST,];
  XI=X[FIRST:LAST,];
  ZI=Z[FIRST:LAST,];
  BETAI_I=B082152[FIRST,]*SELECT`*S082152;
  RESID_I=RESID_I+ZI*BETAI_I`;
  DI=_WEIGHT_[FIRST:LAST,];
  DI=INV(DIAG(DI));
  VI = ZI*PSI_MLE*ZI` + SIG_MLE*DI;
  LNDETERM=LOGDET(VI);
  WI=INVERSE(VI,PSI_MLE,SIG_MLE);
  XWX = XWX + XI`*WI*XI;
  XWY = XWY + XI`*WI*RESID_I;
  YWY = YWY + RESID_I`*WI*RESID_I;
  NCUM=NCUM+RI;
 END;
  ALPH = GINV(XWX)*XWY;
  VAR_MLE = GINV(XWX);
  RSS=YWY-ALPH`*XWY;
  Q_SS=YWY;
  OBJECT=LNDETERM+RSS;
  M2LNL0=M2LNL;
  M2LNL=TOTAL*LOG(2*3.14159) + OBJECT;
 TEMP1=0;TEMP2=0;NCUM=0;
 DO J=1 TO N;
  RI=SAMSIZ[J,];
  FIRST=NCUM+1;LAST=NCUM+RI;
  RESID_I=RESID[FIRST:LAST,];
  XI=X[FIRST:LAST,];
  ZI=Z[FIRST:LAST,];
  BETAI_I=B082152[FIRST,]*SELECT`*S082152;
  RESID_I=RESID_I+ZI*BETAI_I`;
  DI=_WEIGHT_[FIRST:LAST,];
  DI=INV(DIAG(DI));
  VI = ZI*PSI_MLE*ZI` + SIG_MLE*DI;
  WI=INVERSE(VI,PSI_MLE,SIG_MLE);
  BI_MLE = PSI_MLE*ZI`*WI*(RESID_I-XI*ALPH);
  BB082152[FIRST:LAST,]=REPEAT(BI_MLE`*SELECT*S082152,RI,1);
  BI_HAT[FIRST:LAST,] = REPEAT(BI_MLE`,RI,1);
  EI_MLE = SIG_MLE*DI*WI*(RESID_I-XI*ALPH);
  TEMP2=TEMP2+PSI_MLE-PSI_MLE*ZI`*WI*ZI*PSI_MLE+BI_MLE*BI_MLE`;
  TEMP1=TEMP1+SIG_MLE*TRACE(I(RI)-SIG_MLE*DI*WI)+EI_MLE`*INV(DI)*EI_MLE;
  NCUM=NCUM+RI;
 END;
 PSI_MLE=TEMP2/N;
 SIG_MLE=TEMP1/TOTAL;
 SIG_MLE=(1-SIGSQIND)*SIG_MLE + SIGSQIND*SIGSQ;
END;
 B082152=BB082152;
 FINISH;

START RML;         /* REML MODULE FOR FINDING REML ESTIMATES USING THE REML */
                   /* E-M ALGORITHM NESTED WITHIN A LEAST SQUARES LOOP      */
  N_    =N;
  RMIN_ =RMIN;
  RMAX_ =RMAX;
  RAVG_ =RAVG;
  TOTAL_=TOTAL;

/* INITIALIZE ALPHA0, PSI AND SIGMA2 TO BE THE EGLS ESTIMATORS */
 ALPH_RML=ALPH_OLS;
 PSI_RML=PSI_GLS  ;
 SIG_RML=SIG      ;
/*                  END OF INITIALIZATION                    */
 ADJUST=ADJUSTGL;
 M2LNL=1000;M2LNL1=0;
CYCLE=1;
ITER=0;
STEP=0;
SUMMARY=CYCLE||ITER||STEP||ALPH_RML`||M2LNL;
PRINT1="NOTE: CONVERGENCE CRITERION MET.";
PRINT2=" ";
DO ITER=1 TO MAXITER WHILE(ABS(M2LNL-M2LNL1)>CONVERGE);
 CYCLE=2;
 ALPH_EST=ALPH_RML;
 M2LNL1=M2LNL;
 RUN EM_RML;
 ALPH_RML=ALPH_RML+ALPH;
 STEP=SUBITERN-1;
SUMMARY=SUMMARY//(CYCLE||ITER||STEP||ALPH_RML`||M2LNL);
  IF ITER=MAXITER THEN DO;
   PRINT1="WARNING: THE MAXIMUM NUMBER OF ITERATIONS ALLOWED HAVE BEEN  ";
   PRINT2="         REACHED. CONVERGENCE IS ASSUMED.                    ";
  END;
END;
ALPH_EST=ALPH_RML;
PSI_EST=PSI_RML;
SIG_EST=SIG_RML;
RUN SSTOTAL;
ALPHA=ALPH_RML;
PSI=PSI_RML;
SIGPOOL=SIG_RML;
OMEGA=VAR_RML;
RUN ROBUST;
IF COVAR={"ROBUST"} THEN OMEGA=OMEGA_R;
RUN PRINTER;
IF WARNING1=0 THEN DO;
 IF WARNING=0 THEN DO;
  RUN GMANOVA;
  RUN GMANOVAI;
 END;
 ELSE DO;
 PRINT / 'WARNING: THE GMANOVA PROCEDURE COULD NOT BE EXECUTED. THE    ',
         '         CLASS VARIABLE MUST BE SPECIFIED AS THE FIRST       ',
         '         INDEPENDENT VARIABLE LISTED IN THE DESIGN STATEMENT.';
 END;
END;

ELSE DO;
 PRINT / "WARNING: THE OVERALL DESIGN MATRIX SPECIFIED BY THE ALL = STATEMENT",
         "         IS NOT EQUIVALENT ACROSS EACH PARAMETER, B1,B2,...  THERE-",
         "         FORE, ONLY USER-SPECIFIED TESTS OF HYPOTHESES USING THE   ",
         "         TEST STATEMENTS WILL BE PERFORMED.                        ";
END;

IF NO_E>1 THEN DO;
 RUN FUNCTION;
END;
IF NO_C>1 THEN DO;
 RUN CONTRAST;
END;

IF ANY(OPTIONS="OUT") THEN DO;
 RUN OUTPUT;
END;
FINISH;

START EM_RML;      /* THE REML E-M ALGORITHM FOR LINEAR MIXED-EFFECTS MODELS */

 RUN MODEL;
 RUN RANDOM;
 SIG_TEMP=SIG_RML;
 RESID=Y082152-F;
 BB082152 = B082152;
   M2LNL=1000;M2LNL0=0;
 DO SUBITERN=1 TO SUBITER WHILE(ABS(M2LNL-M2LNL0)>CONVERGE);
  XWX=0;XWY=0;YWY=0;LNDETERM=0;NCUM=0;
 DO J=1 TO N;
  RI=SAMSIZ[J,];
  FIRST=NCUM+1;LAST=NCUM+RI;
  RESID_I=RESID[FIRST:LAST,];
  XI=X[FIRST:LAST,];
  ZI=Z[FIRST:LAST,];
  BETAI_I=B082152[FIRST,]*SELECT`*S082152;
  RESID_I=RESID_I+ZI*BETAI_I`;
  DI=_WEIGHT_[FIRST:LAST,];
  DI=INV(DIAG(DI));
  VI = ZI*PSI_RML*ZI` + SIG_RML*DI;
  LNDETERM=LOGDET(VI);
  WI=INVERSE(VI,PSI_RML,SIG_RML);
  XWX = XWX + XI`*WI*XI;
  XWY = XWY + XI`*WI*RESID_I;
  YWY = YWY + RESID_I`*WI*RESID_I;
  NCUM=NCUM+RI;
 END;
  ALPH = GINV(XWX)*XWY;
  VAR_RML  = GINV(XWX);
  RSS=YWY-ALPH`*XWY;
  Q_SS=YWY;
  OBJECT=LNDETERM+RSS;

* New statement which tries to use correct components of XWX when RESTRICT
  is in effect.  The old statements are * out;
  SUBXWX=XWX[LOC(VECDIAG(XWX)>0),LOC(VECDIAG(XWX)>0)];
  CONST=LOG_DET(SUBXWX);
  M2LNL0=M2LNL;
  M2LNL=(TOTAL-PQ)*LOG(2*3.14159) + OBJECT + CONST;
 TEMP1=0;TEMP2=0;NCUM=0;

 DO J=1 TO N;
  RI=SAMSIZ[J,];
  FIRST=NCUM+1;LAST=NCUM+RI;
  RESID_I=RESID[FIRST:LAST,];
  XI=X[FIRST:LAST,];
  ZI=Z[FIRST:LAST,];
  BETAI_I=B082152[FIRST,]*SELECT`*S082152;
  RESID_I=RESID_I+ZI*BETAI_I`;
  DI=_WEIGHT_[FIRST:LAST,];
  DI=INV(DIAG(DI));
  VI = ZI*PSI_RML*ZI` + SIG_RML*DI;
  WI=INVERSE(VI,PSI_RML,SIG_RML);
  PI = WI*( I(RI) - XI*GINV(XWX)*XI`*WI );
  BI_RML = PSI_RML*ZI`*WI*(RESID_I-XI*ALPH);
  BB082152[FIRST:LAST,]=REPEAT(BI_RML`*SELECT*S082152,RI,1);
  BI_HAT[FIRST:LAST,] = REPEAT(BI_RML`,RI,1);
  EI_RML = SIG_RML*DI*WI*(RESID_I-XI*ALPH);
  TEMP2=TEMP2+PSI_RML-PSI_RML*ZI`*PI*ZI*PSI_RML+BI_RML*BI_RML`;
  TEMP1=TEMP1+SIG_RML*TRACE(I(RI)-SIG_RML*DI*PI)+EI_RML`*INV(DI)*EI_RML;
  NCUM=NCUM+RI;
 END;

 PSI_RML=TEMP2/N;
 SIG_RML=TEMP1/TOTAL;
 SIG_RML=(1-SIGSQIND)*SIG_RML + SIGSQIND*SIGSQ;
 END;
 B082152=BB082152;
 FINISH;

START GENNLIN;     /* GENERAL NONLINEAR REGRESSION MODULE DESIGNED TO CALL  */
                   /* THE DIFFERENT MODULES IN THE CORRECT ORDER ACCORDING  */
                   /* TO THE MIXNLIN PROCEDURE OPTIONS.                     */

NAMEB=TRIM(SUBSTR(NAMEB,1,16));
NAMEX=TRIM(SUBSTR(NAMEX,1,30));
VALUE=TRIM(SUBSTR(VALUE,1,30));
NAMEB  = {" ","PARAMETER"," "}//NAMEB;
NAMEX  = {" ","EFFECT"," "}//NAMEX;
VALUE  = {" "," "," "}//VALUE;

RUN GRID;
ALPH_EST=ALPHA0;
RUN MODEL;
RUN RANDOM;
   MLE_IND=0;
   V0=NCOL(Z);
   PQ=NCOL(X)-NO_RSCT;
   PSI0=I(V0);SIG0=1;RANDOM=0;NLITER=OLSITER;
RUN NLIN;

   IF METHOD={"OLS"} THEN GOTO METHOD0;
   IF METHOD={"EGLS"} THEN GOTO METHOD1;
   IF METHOD={"MLE"}  THEN GOTO METHOD2;
   IF METHOD={"REML"} THEN GOTO METHOD3;
   IF METHOD={"IRGLS"} THEN GOTO METHOD4;

METHOD0:
   IF WARNING1=0 THEN DO;
    IF WARNING=0 THEN DO;
     RUN GMANOVA;
     RUN GMANOVAI;
    END;
    ELSE DO;
    PRINT / 'WARNING: THE GMANOVA PROCEDURE COULD NOT BE EXECUTED. THE    ',
            '         CLASS VARIABLE MUST BE SPECIFIED AS THE FIRST       ',
            '         INDEPENDENT VARIABLE LISTED IN THE DESIGN STATEMENT.';
    END;
   END;
   ELSE DO;
    PRINT / "WARNING: THE OVERALL DESIGN MATRIX SPECIFIED BY THE ALL = STATEMENT",
            "         IS NOT EQUIVALENT ACROSS EACH PARAMETER, B1,B2,...  THERE-",
            "         FORE, ONLY USER-SPECIFIED TESTS OF HYPOTHESES USING THE   ",
            "         TEST STATEMENTS WILL BE PERFORMED.                        ";
   END;
   IF NO_E>1 THEN DO;
     RUN FUNCTION;
   END;
   IF NO_C>1 THEN DO;
    RUN CONTRAST;
   END;

   IF ANY(OPTIONS="OUT") THEN DO;
    RUN OUTPUT;
   END;

RETURN;

METHOD1:
   RUN MOMENTS;
   ALPHA0=ALPH_OLS;PSI0=(1-PSI_IND)*PSI_GLS + PSI_IND*PSI_STRT;
   SIG0=SIG;RANDOM=1;NLITER=MAXITER;
   RUN NLIN;
RETURN;

METHOD2:
   RUN MOMENTS;
   ALPHA0=ALPH_OLS;PSI0=(1-PSI_IND)*PSI_MLE + PSI_IND*PSI_STRT;
   SIG0=SIG;RANDOM=1;ADJUSTN=NAME(UNADJUSTED);
   RUN MLE;
RETURN;

METHOD3:
   RUN MOMENTS;
   ALPHA0=ALPH_OLS;PSI0=(1-PSI_IND)*PSI_GLS + PSI_IND*PSI_STRT;
   SIG0=SIG;RANDOM=1;ADJUSTN=NAME(UNADJUSTED);
   RUN RML;
RETURN;

METHOD4:
  RUN MOMENTS;
  ALPHA0=ALPH_OLS;PSI0=(1-PSI_IND)*PSI_GLS + PSI_IND*PSI_STRT;
  SIG0=SIG;RANDOM=1;NLITER=OLSITER;
  RUN IRGLS;
RETURN;

FINISH;

RUN GENNLIN;

FINISH GENMEM;

STORE MODULE=(GENMEM GRID ESTIMATE SEARCH SSTOTAL MOMENTS PRINTER NLIN
              MLE EM_MLE RML EM_RML IRGLS GENNLIN GMANOVA GMANOVAI OUTPUT
              CONTRAST LOGDET LOG_DET INVERSE CONCORR ROBUST FUNCTION
              AIC ESTIM_BI PRED_SS) ;
