
/*                   MIXNLIN (Macro version 3.1)                  */
/*  New Features: 1. SS IRGLS Based on Shrinkage Estimators       */
/*                2. Includes a VARPARMS statement                */
/*                3. Includes an OUTPRED statement outputting     */
/*                   observed and predicted values.               */
/*                4. Includes an ESTIMATE statement               */
/*                5. Change PNMEM F-TEST to MANOVA F-TEST         */ 
/*                6. Put a CYCLE Option in for EGLS               */           
/*                7. Corrected STDERR statement (it was COVAR=)   */ 
/*                8. New Concordance Corr based on Omega          */
/*                9. New RESTRICT Statement for prop. errors      */

/*  Procedure MIXNLIN is a SAS based Macro/IML procedure written  */
/*  explicitly for applications involving nonlinear mixed_effects */
/*  models.  MIXNLIN was written and tested by:                   */
/*                 Edward F. Vonesh, Ph.D.                        */
/*                 Applied Statistics Center                      */
/*                 Baxter Healthcare Corporation                  */
/*                 Route 120 & Wilson Rd.                         */
/*                 Round Lake, IL 60073 USA                       */
/*  Presented below are Macro Statements which must be called up  */
/*  using the %INCLUDE statement (see user manual).               */

%MACRO MIXNLIN(IMLLIB=SASUSER.IMLSTOR);
%GLOBAL STORAGE P DEPEN INDEPEN FUNCTION FUNC CLASS_ BYVAR WT
        OPTIONS METHOD CONVERGE MAXITER USER_I 
        DERB1 DERB2 DERB3 DERB4 DERB5 DERB6
        DERB7 DERB8 DERB9 DERB10 DERB11 DERB12
        USER1 USER2 USER3 USER4 USER5 USER6 USER7 USER8 USER9 USER10 
        USER11 USER12 USER13 USER14 USER15 USER16 USER17 USER18 USER19 USER20 
        USER21 USER22 USER23 USER24 USER25 USER26 USER27 USER28 USER29 USER30
        DSGNB1 DSGNB2 DSGNB3 DSGNB4 DSGNB5 DSGNB6
        DSGNB7 DSGNB8 DSGNB9 DSGN10 DSGN11 DSGN12;
%LET STORAGE=&IMLLIB;
PROC IML;
RESET STORAGE=&STORAGE;
REMOVE _ALL_;
RUN;

OPTIONS IMPLMAC;
LABEL ;
CLASS .;
VARPARMS PSI= ;
WEIGHT 'J(TOTAL,1,1)';
RESTRICT 'NONE';

PUT "***** MIXNLIN 3.1 - 1995 Copyright (c) by Edward F. Vonesh *****";

%LET USER_I=0;
%MEND MIXNLIN;

%MACRO PROCM(MIX,DATA=_LAST_,CONVERGE=.00001,BEST=0,SUBITER=5,OUTPRED=.,
       SIGSQ=-1,OLSITER=50,MAXITER=50,METHOD=EGLS,OPTIONS=GMANOVA,TEST=MODIFIED,
       ALPHA=.05,PRINT=OLS,STEPS=HALVE,EXPAND=AVERAGE,STDERR=ASYMP,
       CYCLE=1,SSMETHOD=EB) / STMT;
%GLOBAL DATA_ _OUT_ OUTPRED1;
%LET DATA_=&DATA;
%IF &OUTPRED=. %THEN %LET OUTPRED=;
%IF &OUTPRED NE %THEN %DO;
%LET _OUT_=OUT;
%LET OUTPRED1=&OUTPRED;
%END;
%ELSE %DO;
%LET _OUT_=;
%LET OUTPRED1=;
%END;

 PROC IML;
 START PROCM;
 TYPE={&TEST};        /* CHOOSE SMALL F-TEST APPROXIMATION METHOD */
 METHOD={&METHOD};    /* SELECT FROM OLS EGLS,IRGLS,MLE AND REML */
 OPTIONS={&PRINT}||{&OPTIONS}||{"&_OUT_"};
                      /* POSSIBLE OPTIONS ARE: SHORT, OLS, ALL */
 CONVERGE=&CONVERGE;  /* CONVERGENCE CRITERIA FOR EGLS,IRGLS,MLE,REML */
 OLSITER =&OLSITER;   /* ITERATIONS FOR OLS (=1 FOR SPECIFIED STARTING VALUES */
 MAXITER =&MAXITER;   /* ITERATIONS FOR LEAST SQUARES LOOP */
 SUBITER =&SUBITER;   /* SUBITERATIONS FOR E-M ALGORITHMS  */
 NCYCLE  =&CYCLE;     /* NUMBER OF CYCLES OF EGLS */ 
 BEST=&BEST;          /* SELECT BEST N RESIDUAL SS */
 STEPS={&STEPS};      /* HALVE OR NOHALVE, SELECT NOHALVE IF WEIGHT
                         STATEMENT HAS VALUES DEPENDING ON PARMS */
 EXPAND={&EXPAND};    /* EXPAND=AVERAGE(SHEINER) OR CURRENT(LINDSTROM) */
 SSMETHOD={&SSMETHOD};/* SSMETHOD=EB(EMPIRICAL BAYES) OR LS(LEAST SQUARES) */ 
 COVAR={&STDERR};     /* COVAR=ROBUST OR ASYMP FOR CALCULATING STD ERRORS */
 TYPE_I=&ALPHA;       /* TYPE I ERROR (ALPHA) FOR CONFIDENCE INTERVALS */
 SIGSQ=&SIGSQ;        /* SIGSQ=-1 (DEFAULT) OR USER'S RESTRICTED VALUE */
 SIG82152=(SIGSQ=0);  /* SIG82152=0 OR 1. SET SIGSQ=0 (SIG82152=1) IF WE WISH 
                         TO ALLOW RI=V0 FOR UNRESTRICTED MANOVA MODELS */
IF COVAR={"ASYMP"} THEN COVAR={"ASYMPTOTIC"};
IF METHOD={"MLE"}  THEN SSMETHOD={"EB"};
IF METHOD={"REML"} THEN SSMETHOD={"EB"};

 FINISH;
 RUN PROCM;
 RESET STORAGE=&STORAGE;
 STORE METHOD OPTIONS CONVERGE OLSITER MAXITER SUBITER NCYCLE 
       BEST TYPE STEPS EXPAND SSMETHOD COVAR TYPE_I SIG82152 SIGSQ;
%MEND PROCM;

%MACRO SUBJECT(_ID_) / STMT;
%GLOBAL BYVAR;
 %LET BYVAR=&_ID_;
%MEND SUBJECT;

%MACRO CLASS(_CLASS_) / STMT;
%GLOBAL CLASS_;
%IF &_CLASS_=. %THEN %LET _CLASS_=;
%IF &_CLASS_ NE %THEN %DO;
 %LET CLASS_=&_CLASS_;
%END;
%ELSE %DO;
 %LET CLASS_=;
%END;
%MEND CLASS;

%MACRO VARS(_VAR_) / STMT;
DATA;
 INDEPEN=TRANSLATE(&_VAR_,'##','**');
 CALL SYMPUT('INDEPEN',INDEPEN);
%MEND VARS;

/* MACRO INDEP ALLOWS US TO TEMPORARILY CALL UP INDEPENDENT VARIABLES IN    */
/* VARIOUS PARTS OF THE PROGRAM AND STILL ALLOW IML MATRICES TO BE NAMED AS */
/* THEY ARE WITHOUT CONFLICT BETWEEN VARIABLES LISTED IN THE VARS STATEMENT */
/* AND IML MATRICES AS I HAVE LABELLED THEM INSIDE THE VARIOUS IML MODULES. */

%MACRO INDEP; 
 %LET I=0;
 %DO %WHILE(%SCAN(&INDEPEN,&I+1,%STR( )) NE);
 %LET I=%EVAL(&I+1);
 %SCAN(&INDEPEN,&I,%STR( ))=INDEP[,&I];
 %END;
%MEND;

/* MACRO PARMS IS DESIGNED TO GET AN ARRAY OF STARTING VALUES TO BE USED */
/* IN A GRID SEARCH USING THE IML SUBROUTINE: GRID.                      */

%MACRO PARMS(B1=,B2=,B3=,B4=,B5=,B6=,B7=,B8=,B9=,B10=,B11=,B12=) / STMT;

DATA;
RETAIN ERRORB 0;
ERRORA=0;
ERROR ='ERROR: THE FIRST PARAMETER, B1, HAS NOT BEEN INITIALIZED';
ERROR1='ERROR: AT LEAST ONE PARAMETER HAS BEEN SPECIFIED OUT OF ORDER.';
ERROR2='       PARAMETERS MUST BE SPECIFIED IN THE ORDER B1,B2,B3,... ';
%LET P=0;
%IF &B1 NE %THEN %DO;%LET P=%EVAL(&P+1);
B1=TRANSLATE(&B1,'##','**');
CALL SYMPUT('B1',B1);
%END;
%ELSE %DO;ERRORA=1;ERRORB=1;%END;
%IF &B2 NE AND &P<1 %THEN %DO;ERRORB=1;%END;
%IF &B2 NE AND &B1 NE %THEN %DO;%LET P=%EVAL(&P+1);
B2=TRANSLATE(&B2,'##','**');
CALL SYMPUT('B2',B2);
%END;
%IF &B3 NE AND &P<2 %THEN %DO;ERRORB=1;%END;
%IF &B3 NE AND &B2 NE AND &B1 NE %THEN %DO;%LET P=%EVAL(&P+1);
B3=TRANSLATE(&B3,'##','**');
CALL SYMPUT('B3',B3);
%END;
%IF &B4 NE AND &P<3 %THEN %DO;ERRORB=1;%END;
%IF &B4 NE AND &B3 NE AND &B2 NE AND &B1 NE
%THEN %DO;%LET P=%EVAL(&P+1);
B4=TRANSLATE(&B4,'##','**');
CALL SYMPUT('B4',B4);
%END;
%IF &B5 NE AND &P<4 %THEN %DO;ERRORB=1;%END;
%IF &B5 NE AND &B4 NE AND &B3 NE AND &B2 NE AND &B1 NE
%THEN %DO;%LET P=%EVAL(&P+1);
B5=TRANSLATE(&B5,'##','**');
CALL SYMPUT('B5',B5);
%END;
%IF &B6 NE AND &P<5 %THEN %DO;ERRORB=1;%END;
%IF &B6 NE AND &B5 NE AND &B4 NE AND &B3 NE AND &B2 NE AND &B1 NE
%THEN %DO;%LET P=%EVAL(&P+1);
B6=TRANSLATE(&B6,'##','**');
CALL SYMPUT('B6',B6);
%END;
%IF &B7 NE AND &P<6 %THEN %DO;ERRORB=1;%END;
%IF &B7 NE AND &B6 NE AND &B5 NE AND &B4 NE AND &B3 NE AND &B2 NE AND &B1 NE
%THEN %DO;%LET P=%EVAL(&P+1);
B7=TRANSLATE(&B7,'##','**');
CALL SYMPUT('B7',B7);
%END;
%IF &B8 NE AND &P<7 %THEN %DO;ERRORB=1;%END;
%IF &B8 NE AND &B7 NE AND
 &B6 NE AND &B5 NE AND &B4 NE AND &B3 NE AND &B2 NE AND &B1 NE
%THEN %DO;%LET P=%EVAL(&P+1);
B8=TRANSLATE(&B8,'##','**');
CALL SYMPUT('B8',B8);
%END;
%IF &B9 NE AND &P<8 %THEN %DO;ERRORB=1;%END;
%IF &B9 NE AND &B8 NE AND &B7 NE AND
 &B6 NE AND &B5 NE AND &B4 NE AND &B3 NE AND &B2 NE AND &B1 NE
%THEN %DO;%LET P=%EVAL(&P+1);
B9=TRANSLATE(&B9,'##','**');
CALL SYMPUT('B9',B9);
%END;
%IF &B10 NE AND &P<9 %THEN %DO;ERRORB=1;%END;
%IF &B10 NE AND &B9 NE AND &B8 NE AND &B7 NE AND
 &B6 NE AND &B5 NE AND &B4 NE AND &B3 NE AND &B2 NE AND &B1 NE
%THEN %DO;%LET P=%EVAL(&P+1);
B10=TRANSLATE(&B10,'##','**');
CALL SYMPUT('B10',B10);
%END;
%IF &B11 NE AND &P<10 %THEN %DO;ERRORB=1;%END;
%IF &B11 NE AND &B10 NE AND &B9 NE AND &B8 NE AND &B7 NE AND
 &B6 NE AND &B5 NE AND &B4 NE AND &B3 NE AND &B2 NE AND &B1 NE
%THEN %DO;%LET P=%EVAL(&P+1);
B11=TRANSLATE(&B11,'##','**');
CALL SYMPUT('B11',B11);
%END;
%IF &B12 NE AND &P<11 %THEN %DO;ERRORB=1;%END;
%IF &B12 NE AND &B11 NE AND &B10 NE AND &B9 NE AND &B8 NE AND &B7 NE AND
 &B6 NE AND &B5 NE AND &B4 NE AND &B3 NE AND &B2 NE AND &B1 NE
%THEN %DO;%LET P=%EVAL(&P+1);
B12=TRANSLATE(&B12,'##','**');
CALL SYMPUT('B12',B12);
%END;

IF ERRORA=1 THEN DO;
 PUT ERROR;
END;
IF ERRORB=1 THEN DO;
 PUT ERROR1 ERROR2;
END;

PROC IML;
RESET STORAGE=&STORAGE;
LOAD LABELID;
START PARMS;
 P=&P;
 PQ=0;
 PARMID={B1};
%DO I=2 %TO &P;
 PARMID=PARMID||{B&I};
%END;
%DO I=1 %TO &P;
 B&I={&&&B&I};
 PQ=PQ+NCOL(B&I);
%END;

 TEMP=B1;TEMP1=B1;TEMPLAST=B1;
%DO I=2 %TO &P;
NROWB=NROW(B&I);
 TEMP=TEMPLAST || ( J(NROW(TEMPLAST),1,1)@B&I[1,] );
DO K=2 TO NROWB;
 TEMP1=TEMPLAST || ( J(NROW(TEMPLAST),1,1)@B&I[K,] );
 TEMP=TEMP//TEMP1;
END;
 TEMPLAST=TEMP;
%END;
 ALPHA0=TEMP;
 LABELID=LABELID[,(1:&P)];
FINISH;
RUN PARMS;
RESET STORAGE=&STORAGE;
STORE LABELID PARMID ALPHA0 PQ;
%TEST;
%ESTIMATE;
%MEND PARMS;

%MACRO VARPARMS(PSI=) / STMT;
 %LET PSILST=%QSCAN(&PSI,1,%STR(%')=);

PROC IML;
RESET STORAGE=&STORAGE;
START GETPSI; 
PSI_IND=0;
PSI_STRT=1;
%IF &PSILST NE %THEN %DO;
PSI_IND=1;
PSI_STRT=&PSILST;
%END;
STORE PSI_IND PSI_STRT;
FINISH;
RUN GETPSI;
%MEND VARPARMS;

%MACRO LABEL(B1='B1',B2='B2',B3='B3',B4='B4',B5='B5',B6='B6',
             B7='B7',B8='B8',B9='B9',B10='B10',B11='B11',B12='B12') / STMT;
PROC IML;
START LABELS;
LABELID={&B1};
%DO I=2 %TO 12;
LABELID=LABELID||{&&&B&I};
%END;
FINISH;
RUN LABELS;
RESET STORAGE=&STORAGE;
STORE LABELID;
%MEND LABEL;

%MACRO MODEL(MACMODEL) / STMT;
%GLOBAL DEPEN FUNCTION FUNC;
 %LET DEPEN=%SCAN(&MACMODEL,1,%STR(%')=);
 %LET FUNCTION=%SCAN(&MACMODEL,2,%STR(%')=);
 DATA;
 FUNC=TRANSLATE("&FUNCTION",'##','**');
 CALL SYMPUT('FUNC',FUNC);
%MEND MODEL;

%MACRO DERIV(DER_B1=,DER_B2=,DER_B3=,DER_B4=,DER_B5=,DER_B6=,DER_B7=,
             DER_B8=,DER_B9=,DER_B10=,DER_B11=,DER_B12=) / STMT;
DATA;
%IF &DER_B1 NE %THEN %DO;
DERB1=TRANSLATE(&DER_B1,'##','**');
CALL SYMPUT('DERB1',DERB1);
%END;
%IF &DER_B2 NE %THEN %DO;
DERB2=TRANSLATE(&DER_B2,'##','**');
CALL SYMPUT('DERB2',DERB2);
%END;
%IF &DER_B3 NE %THEN %DO;
DERB3=TRANSLATE(&DER_B3,'##','**');
CALL SYMPUT('DERB3',DERB3);
%END;
%IF &DER_B4 NE %THEN %DO;
DERB4=TRANSLATE(&DER_B4,'##','**');
CALL SYMPUT('DERB4',DERB4);
%END;
%IF &DER_B5 NE %THEN %DO;
DERB5=TRANSLATE(&DER_B5,'##','**');
CALL SYMPUT('DERB5',DERB5);
%END;
%IF &DER_B6 NE %THEN %DO;
DERB6=TRANSLATE(&DER_B6,'##','**');
CALL SYMPUT('DERB6',DERB6);
%END;
%IF &DER_B7 NE %THEN %DO;
DERB7=TRANSLATE(&DER_B7,'##','**');
CALL SYMPUT('DERB7',DERB7);
%END;
%IF &DER_B8 NE %THEN %DO;
DERB8=TRANSLATE(&DER_B8,'##','**');
CALL SYMPUT('DERB8',DERB8);
%END;
%IF &DER_B9 NE %THEN %DO;
DERB9=TRANSLATE(&DER_B9,'##','**');
CALL SYMPUT('DERB9',DERB9);
%END;
%IF &DER_B10 NE %THEN %DO;
DERB10=TRANSLATE(&DER_B10,'##','**');
CALL SYMPUT('DERB10',DERB10);
%END;
%IF &DER_B11 NE %THEN %DO;
DERB11=TRANSLATE(&DER_B11,'##','**');
CALL SYMPUT('DERB11',DERB11);
%END;
%IF &DER_B12 NE %THEN %DO;
DERB12=TRANSLATE(&DER_B12,'##','**');
CALL SYMPUT('DERB12',DERB12);
%END;
%MEND DERIV;

%MACRO CALLDERB;
  DERB1=&DERB1;
  DERB=DERB1;
  %DO I=2 %TO &P;
  DERB&I=&&&DERB&I;
  DERB=DERB||(DERB&I);
  %END;
%MEND CALLDERB;

%MACRO FUNCTION(USER_1=,USER_2=,USER_3=,USER_4=,USER_5=,USER_6=,USER_7=,
                USER_8=,USER_9=,USER_10=,USER_11=,USER_12=,
                USER_13=,USER_14=,USER_15=,USER_16=,USER_17=,USER_18=,
                USER_19=,USER_20=,USER_21=,USER_22=,USER_23=,USER_24=,
                USER_25=,USER_26=,USER_27=,USER_28=,USER_29=,USER_30=) /STMT;
DATA;
%IF &USER_1 NE %THEN %DO;
%LET USER_I=%EVAL(&USER_I+1);
USER1=TRANSLATE(&USER_1,'##','**');
CALL SYMPUT('USER1',USER1);
%END;
%IF &USER_2 NE %THEN %DO;
%LET USER_I=%EVAL(&USER_I+1);
USER2=TRANSLATE(&USER_2,'##','**');
CALL SYMPUT('USER2',USER2);
%END;
%IF &USER_3 NE %THEN %DO;
%LET USER_I=%EVAL(&USER_I+1);
USER3=TRANSLATE(&USER_3,'##','**');
CALL SYMPUT('USER3',USER3);
%END;
%IF &USER_4 NE %THEN %DO;
%LET USER_I=%EVAL(&USER_I+1);
USER4=TRANSLATE(&USER_4,'##','**');
CALL SYMPUT('USER4',USER4);
%END;
%IF &USER_5 NE %THEN %DO;
%LET USER_I=%EVAL(&USER_I+1);
USER5=TRANSLATE(&USER_5,'##','**');
CALL SYMPUT('USER5',USER5);
%END;
%IF &USER_6 NE %THEN %DO;
%LET USER_I=%EVAL(&USER_I+1);
USER6=TRANSLATE(&USER_6,'##','**');
CALL SYMPUT('USER6',USER6);
%END;
%IF &USER_7 NE %THEN %DO;
%LET USER_I=%EVAL(&USER_I+1);
USER7=TRANSLATE(&USER_7,'##','**');
CALL SYMPUT('USER7',USER7);
%END;
%IF &USER_8 NE %THEN %DO;
%LET USER_I=%EVAL(&USER_I+1);
USER8=TRANSLATE(&USER_8,'##','**');
CALL SYMPUT('USER8',USER8);
%END;
%IF &USER_9 NE %THEN %DO;
%LET USER_I=%EVAL(&USER_I+1);
USER9=TRANSLATE(&USER_9,'##','**');
CALL SYMPUT('USER9',USER9);
%END;
%IF &USER_10 NE %THEN %DO;
%LET USER_I=%EVAL(&USER_I+1);
USER10=TRANSLATE(&USER_10,'##','**');
CALL SYMPUT('USER10',USER10);
%END;
%IF &USER_11 NE %THEN %DO;
%LET USER_I=%EVAL(&USER_I+1);
USER11=TRANSLATE(&USER_11,'##','**');
CALL SYMPUT('USER11',USER11);
%END;
%IF &USER_12 NE %THEN %DO;
%LET USER_I=%EVAL(&USER_I+1);
USER12=TRANSLATE(&USER_12,'##','**');
CALL SYMPUT('USER12',USER12);
%END;
%IF &USER_13 NE %THEN %DO;
%LET USER_I=%EVAL(&USER_I+1);
USER13=TRANSLATE(&USER_13,'##','**');
CALL SYMPUT('USER13',USER13);
%END;
%IF &USER_14 NE %THEN %DO;
%LET USER_I=%EVAL(&USER_I+1);
USER14=TRANSLATE(&USER_14,'##','**');
CALL SYMPUT('USER14',USER14);
%END;
%IF &USER_15 NE %THEN %DO;
%LET USER_I=%EVAL(&USER_I+1);
USER15=TRANSLATE(&USER_15,'##','**');
CALL SYMPUT('USER15',USER15);
%END;
%IF &USER_16 NE %THEN %DO;
%LET USER_I=%EVAL(&USER_I+1);
USER16=TRANSLATE(&USER_16,'##','**');
CALL SYMPUT('USER16',USER16);
%END;
%IF &USER_17 NE %THEN %DO;
%LET USER_I=%EVAL(&USER_I+1);
USER17=TRANSLATE(&USER_17,'##','**');
CALL SYMPUT('USER17',USER17);
%END;
%IF &USER_18 NE %THEN %DO;
%LET USER_I=%EVAL(&USER_I+1);
USER18=TRANSLATE(&USER_18,'##','**');
CALL SYMPUT('USER18',USER18);
%END;
%IF &USER_19 NE %THEN %DO;
%LET USER_I=%EVAL(&USER_I+1);
USER19=TRANSLATE(&USER_19,'##','**');
CALL SYMPUT('USER19',USER19);
%END;
%IF &USER_20 NE %THEN %DO;
%LET USER_I=%EVAL(&USER_I+1);
USER20=TRANSLATE(&USER_20,'##','**');
CALL SYMPUT('USER20',USER20);
%END;
%IF &USER_21 NE %THEN %DO;
%LET USER_I=%EVAL(&USER_I+1);
USER21=TRANSLATE(&USER_21,'##','**');
CALL SYMPUT('USER21',USER21);
%END;
%IF &USER_22 NE %THEN %DO;
%LET USER_I=%EVAL(&USER_I+1);
USER22=TRANSLATE(&USER_22,'##','**');
CALL SYMPUT('USER22',USER22);
%END;
%IF &USER_23 NE %THEN %DO;
%LET USER_I=%EVAL(&USER_I+1);
USER23=TRANSLATE(&USER_23,'##','**');
CALL SYMPUT('USER23',USER23);
%END;
%IF &USER_24 NE %THEN %DO;
%LET USER_I=%EVAL(&USER_I+1);
USER24=TRANSLATE(&USER_24,'##','**');
CALL SYMPUT('USER24',USER24);
%END;
%IF &USER_25 NE %THEN %DO;
%LET USER_I=%EVAL(&USER_I+1);
USER25=TRANSLATE(&USER_25,'##','**');
CALL SYMPUT('USER25',USER25);
%END;
%IF &USER_26 NE %THEN %DO;
%LET USER_I=%EVAL(&USER_I+1);
USER26=TRANSLATE(&USER_26,'##','**');
CALL SYMPUT('USER26',USER26);
%END;
%IF &USER_27 NE %THEN %DO;
%LET USER_I=%EVAL(&USER_I+1);
USER27=TRANSLATE(&USER_27,'##','**');
CALL SYMPUT('USER27',USER27);
%END;
%IF &USER_28 NE %THEN %DO;
%LET USER_I=%EVAL(&USER_I+1);
USER28=TRANSLATE(&USER_28,'##','**');
CALL SYMPUT('USER28',USER28);
%END;
%IF &USER_29 NE %THEN %DO;
%LET USER_I=%EVAL(&USER_I+1);
USER29=TRANSLATE(&USER_29,'##','**');
CALL SYMPUT('USER29',USER29);
%END;
%IF &USER_30 NE %THEN %DO;
%LET USER_I=%EVAL(&USER_I+1);
USER30=TRANSLATE(&USER_30,'##','**');
CALL SYMPUT('USER30',USER30);
%END;

%MEND FUNCTION;


%MACRO CALLUSER;
%IF &USER_I>0 %THEN %DO;
  %DO I=1 %TO &USER_I;
  USER_&I=&&&USER&I;
  %END;
%END;
%MEND CALLUSER;

%MACRO RANDOM(_RANDOM_) / STMT;
%GLOBAL RANDOM_;
DATA;
 RANDOM_=TRANSLATE(&_RANDOM_,'##','**');
 CALL SYMPUT('RANDOM_',RANDOM_);
%MEND RANDOM;

%MACRO DESIGN(ALL=,B1=,B2=,B3=,B4=,B5=,B6=,B7=,B8=,B9=,B10=,B11=,B12=) / STMT;
%GLOBAL DSGNALL MESSAGE WARNING;
 %LET MESSAGE=0;
%IF &ALL EQ %THEN %DO;
 %LET MESSAGE=1;
%END;

 DATA;
%IF &ALL NE %THEN %DO;
  DSGNALL=TRANSLATE(&ALL,'##','**');
  CALL SYMPUT('DSGNALL',DSGNALL);
%END;
  %IF &B1 NE %THEN %DO;
    DSGNB1=TRANSLATE(&B1,'##','**');
    CALL SYMPUT('DSGNB1',DSGNB1);
  %END;
  %ELSE %DO;
    CALL SYMPUT('DSGNB1',DSGNALL);
  %END;
  %IF &B2 NE %THEN %DO;
    DSGNB2=TRANSLATE(&B2,'##','**');
    CALL SYMPUT('DSGNB2',DSGNB2);
  %END;
  %ELSE %DO;
    CALL SYMPUT('DSGNB2',DSGNALL);
  %END;
  %IF &B3 NE %THEN %DO;
    DSGNB3=TRANSLATE(&B3,'##','**');
    CALL SYMPUT('DSGNB3',DSGNB3);
  %END;
  %ELSE %DO;
    CALL SYMPUT('DSGNB3',DSGNALL);
  %END;
  %IF &B4 NE %THEN %DO;
    DSGNB4=TRANSLATE(&B4,'##','**');
    CALL SYMPUT('DSGNB4',DSGNB4);
  %END;
  %ELSE %DO;
    CALL SYMPUT('DSGNB4',DSGNALL);
  %END;
  %IF &B5 NE %THEN %DO;
    DSGNB5=TRANSLATE(&B5,'##','**');
    CALL SYMPUT('DSGNB5',DSGNB5);
  %END;
  %ELSE %DO;
    CALL SYMPUT('DSGNB5',DSGNALL);
  %END;
  %IF &B6 NE %THEN %DO;
    DSGNB6=TRANSLATE(&B6,'##','**');
    CALL SYMPUT('DSGNB6',DSGNB6);
  %END;
  %ELSE %DO;
    CALL SYMPUT('DSGNB6',DSGNALL);
  %END;
  %IF &B7 NE %THEN %DO;
    DSGNB7=TRANSLATE(&B7,'##','**');
    CALL SYMPUT('DSGNB7',DSGNB7);
  %END;
  %ELSE %DO;
    CALL SYMPUT('DSGNB7',DSGNALL);
  %END;
  %IF &B8 NE %THEN %DO;
    DSGNB8=TRANSLATE(&B8,'##','**');
    CALL SYMPUT('DSGNB8',DSGNB8);
  %END;
  %ELSE %DO;
    CALL SYMPUT('DSGNB8',DSGNALL);
  %END;
  %IF &B9 NE %THEN %DO;
    DSGNB9=TRANSLATE(&B9,'##','**');
    CALL SYMPUT('DSGNB9',DSGNB9);
  %END;
  %ELSE %DO;
    CALL SYMPUT('DSGNB9',DSGNALL);
  %END;
  %IF &B10 NE %THEN %DO;
    DSGNB10=TRANSLATE(&B10,'##','**');
    CALL SYMPUT('DSGNB10',DSGNB10);
  %END;
  %ELSE %DO;
    CALL SYMPUT('DSGNB10',DSGNALL);
  %END;
  %IF &B11 NE %THEN %DO;
    DSGNB11=TRANSLATE(&B11,'##','**');
    CALL SYMPUT('DSGNB11',DSGNB11);
  %END;
  %ELSE %DO;
    CALL SYMPUT('DSGNB12',DSGNALL);
  %END;
  %IF &B12 NE %THEN %DO;
    DSGNB12=TRANSLATE(&B12,'##','**');
    CALL SYMPUT('DSGNB12',DSGNB12);
  %END;
  %ELSE %DO;
    CALL SYMPUT('DSGNB12',DSGNALL);
  %END;

%LET WARNING=0;
%DO I=1 %TO 12;
 %IF &&&B&I NE %THEN %DO;
 %LET WARNING=1;
 %END;
%END;

%MEND DESIGN;

%MACRO RESTRICT(_RESTCT_) / STMT;
%GLOBAL RESTRCT;
DATA;
 RESTRCT=TRANSLATE(&_RESTCT_,'##','**');
 CALL SYMPUT('RESTRCT', RESTRCT);
%MEND RESTRICT;

%MACRO _X2_;
%LOCAL K KK;
  %LET K=1;
  %LET KK=%QSCAN(&DSGNALL,&K,%STR( ));
  NO_CLASS=NCOL(&CLASS_);
  DSGNALL=%QSCAN(&DSGNALL,&K,%STR( ));
  TMPCOLA=NCOL(DSGNALL);
 %DO %UNTIL(&KK EQ);
  %LET K=%EVAL(&K)+1;
  %LET KK=%QSCAN(&DSGNALL,%EVAL(&K+1),%STR( ));
  %IF %QSCAN(&DSGNALL,%EVAL(&K),%STR( )) EQ %THEN %GOTO STOP0;
  DSGNALL=(DSGNALL)||(%QSCAN(&DSGNALL,&K,%STR( )));
  TMPCOLA=TMPCOLA||(NCOL(%QSCAN(&DSGNALL,&K,%STR( ))));

 %END;
 %STOP0:
   NCOLALL=NCOL(DSGNALL);
%MEND _X2_;

%MACRO SETDSGN;
%LOCAL K KK;
%DO I=1 %TO &P;
  %LET K=1;
  %LET KK=%QSCAN(&&&DSGNB&I,&K,%STR( ));
  DSGNB&I=%QSCAN(&&&DSGNB&I,&K,%STR( ));
  TMPCOL&I=NCOL(DSGNB&I);
 %DO %UNTIL(&KK EQ);
  %LET K=%EVAL(&K)+1;
  %LET KK=%QSCAN(&&&DSGNB&I,%EVAL(&K+1),%STR( ));
  %IF %QSCAN(&&&DSGNB&I,%EVAL(&K),%STR( )) EQ %THEN %GOTO STOP1;
  DSGNB&I=(DSGNB&I)||(%QSCAN(&&&DSGNB&I,&K,%STR( )));
  TMPCOL&I=TMPCOL&I||(NCOL(%QSCAN(&&&DSGNB&I,&K,%STR( ))));
 %END;
 %STOP1:
   NCOLB&I=NCOL(DSGNB&I);
 %END;
%MEND SETDSGN;

%MACRO GMANOVA;
%LOCAL K KK;
 %LET K=1;
 %LET KK=%QSCAN(&DSGNALL,&K,%STR( ));
 TESTALL=%QSCAN(&DSGNALL,&K,%STR( ));
 COLLALL=NCOL(TESTALL);
 LABELS={"%QSCAN(&DSGNALL,&K,%STR( ))"};
 IF LABELS={"&CLASS_"} THEN DO;
  NO_COVAR=Q-NO_CLASS;
 %DO %UNTIL(&KK EQ);
  %LET K=%EVAL(&K+1);
  %LET KK=%QSCAN(&DSGNALL,%EVAL(&K+1),%STR( ));
  %IF %QSCAN(&DSGNALL,%EVAL(&K),%STR( )) EQ %THEN %GOTO STOP3;
  TESTALL=%QSCAN(&DSGNALL,&K,%STR( ));
  COLLALL=NCOL(TESTALL);
  IF COLLALL >1 THEN
  LABELS=LABELS//
  ( ROWCAT( ({"%QSCAN(&DSGNALL,&K,%STR( ))"}//J((COLLALL-1),1," ")) ||
            (J(COLLALL,1,{'-'})||DESCRIP0) ) );
  ELSE LABELS=LABELS//({"%QSCAN(&DSGNALL,&K,%STR( ))"});
 %END;
 %STOP3:
  WARNING=0;
 END;

 ELSE DO;
  NO_COVAR=Q;
  WARNING=1;
 END;
 LABELS=TRIM(LABELS);
%MEND GMANOVA;



%MACRO _SOURCE_;
%LOCAL K KK;
%DO I=1 %TO &P;
  %LET K=1;
  %LET KK=%QSCAN(&&&DSGNB&I,&K,%STR( ));
  TEMP&I={"%QSCAN(&&&DSGNB&I,&K,%STR( ))"};
 %DO %UNTIL(&KK EQ);
  %LET K=%EVAL(&K+1);
  %LET KK=%QSCAN(&&&DSGNB&I,%EVAL(&K+1),%STR( ));
  %IF %QSCAN(&&&DSGNB&I,%EVAL(&K),%STR( )) EQ %THEN %GOTO STOP2;
  TEMP&I=(TEMP&I)||{"%QSCAN(&&&DSGNB&I,&K,%STR( ))"};
 %END;
 %STOP2:
 NTEMP&I=NCOL(TEMP&I);
 ALPHAN&I=CONCAT({"B&I"},J(NCOLB&I,1,"."),CHAR((1:NCOLB&I)`,2,0));
 IF NCOLB&I=1 THEN DO; NAMEB&I =LABELID[,&I]; END;
 ELSE DO; NAMEB&I =LABELID[,&I]//J((NCOLB&I - 1),1," "); END;
 NAMET&I=LABELID[,&I];
 NAMEX&I= J(NCOLB&I,1,"                                        ");
 VALUE&I= J(NCOLB&I,1,"                                        ");
 LASTI=0;
  DO J=1 TO NTEMP&I;
   TMPCOL=TMPCOL&I[,J];
   FIRSTI=LASTI+1;LASTI=LASTI+TMPCOL;
   FIRST=FIRSTI;LAST=LASTI;
   NAMEX&I[FIRST,]=TEMP&I[,J];
   IF TMPCOL>1 THEN VALUE&I[FIRST:LAST,]=DESCRIP0;
   ELSE VALUE&I[FIRST:LAST,]=J(TMPCOL,1," ");
  END;
%END;
ALPHAN=COMPRESS(ALPHAN1);
NAMEB=TRIM(NAMEB1);
NAMETEST=TRIM(NAMET1);
NAMEX=TRIM(NAMEX1);
VALUE=TRIM(VALUE1);
%DO I=2 %TO &P;
 ALPHAN=ALPHAN//(COMPRESS(ALPHAN&I));
 NAMEB=NAMEB//(TRIM(NAMEB&I));
 NAMETEST=NAMETEST//(TRIM(NAMET&I));
 NAMEX=NAMEX//(TRIM(NAMEX&I));
 VALUE=VALUE//(TRIM(VALUE&I));
%END;
 ALPHAN=COMPRESS(ALPHAN);
 NAMEB=TRIM(NAMEB);
 NAMETEST=TRIM(NAMETEST);
 NAMEX=TRIM(NAMEX);
 VALUE=TRIM(VALUE);

%MEND _SOURCE_;

%MACRO BETA_F;
LASTI=0;
%DO I=1 %TO &P;
 FIRSTI=LASTI+1;LASTI=LASTI+NCOLB&I;
 FIRST=FIRSTI;LAST=LASTI;
 B&I=(DSGNB&I)*ALPH_EST[FIRST:LAST,1] + B082152[,&I];
%END;
%MEND BETA_F;

%MACRO BETA_W;
LASTI=0;
%DO I=1 %TO &P;
 FIRSTI=LASTI+1;LASTI=LASTI+NCOLB&I;
 FIRST=FIRSTI;LAST=LASTI;
 B&I=(DSGNB&I)*ALPH_EST[FIRST:LAST,1];
%END;
%MEND BETA_W;

%MACRO WEIGHT(_WT_) / STMT;
%GLOBAL WT;
 DATA;
 WT=TRANSLATE(&_WT_,'##','**');
 CALL SYMPUT('WT',WT);
%MEND WEIGHT;

%MACRO XFIX;
 X082152 = DERB[,1]#DSGNB1#BSGN[1,1];
 %DO I=2 %TO &P;
  X082152 = X082152||(DERB[,&I]#DSGNB&I#BSGN[&I,1]);
 %END;
%MEND XFIX;

%MACRO TEST(LABEL=,CONTRAST=) / STMT;
%IF &LABEL EQ %THEN %DO;
PROC IML;
RESET STORAGE=&STORAGE;
LOAD PQ;
START CTRAST0;
CSOURCE='RESIDUAL';
NO_C=1;
RANK_C=0;
CONTRAST=J(1,PQ,0);
FINISH;
RUN CTRAST0;
RESET STORAGE=&STORAGE;
STORE CSOURCE NO_C CONTRAST PQ RANK_C;
%END;


%IF &LABEL NE %THEN %DO;
DATA;
CLABEL=TRANSLATE(&LABEL,'##','**');
CALL SYMPUT('CLABEL',CLABEL);
CTRAST=TRANSLATE(&CONTRAST,'##','**');
CALL SYMPUT('CTRAST',CTRAST);

PROC IML;
RESET STORAGE=&STORAGE;
LOAD CSOURCE NO_C CONTRAST PQ RANK_C;
START CTRAST;
CSOURCE=CSOURCE//{"&CLABEL"};
NO_C=NO_C+1;
RANK_C=RANK_C//NROW({&CTRAST});
CONTRAST=CONTRAST//{&CTRAST};
FINISH;
RUN CTRAST;
STORE CSOURCE NO_C CONTRAST PQ RANK_C;
%END;
%MEND TEST;

%MACRO ESTIMATE(LABEL=,VALUE=) / STMT;
%IF &LABEL EQ %THEN %DO;
PROC IML;
RESET STORAGE=&STORAGE;
LOAD PQ;
START ESTIMAT0;
ESOURCE='RESIDUAL';
NO_E=1;
RANK_E=0;
ESTIMAT=J(1,PQ,0);
FINISH;
RUN ESTIMAT0;
RESET STORAGE=&STORAGE;
STORE ESOURCE NO_E ESTIMAT PQ RANK_E;
%END;


%IF &LABEL NE %THEN %DO;
DATA;
ELABEL=TRANSLATE(&LABEL,'##','**');
CALL SYMPUT('ELABEL',ELABEL);
ETRAST=TRANSLATE(&VALUE,'##','**');
CALL SYMPUT('ETRAST',ETRAST);

PROC IML;
RESET STORAGE=&STORAGE;
LOAD ESOURCE NO_E ESTIMAT PQ RANK_E;
START ETRAST;
ESOURCE=ESOURCE//{"&ELABEL"};
NO_E=NO_E+1;
RANK_E=RANK_E//NROW({&ETRAST});
ESTIMAT=ESTIMAT//{&ETRAST};
FINISH;
RUN ETRAST;
STORE ESOURCE NO_E ESTIMAT PQ RANK_E;
%END;
%MEND ESTIMATE;


%MACRO GETDATA;
 DATA RM82152D;SET &DATA_;
 %IF %LENGTH(%SCAN(&CLASS_,1))=0 %THEN %DO;
  %LET CLASS_=_CLASS_;
  DATA RM82152D;SET RM82152D;
  &CLASS_='OVERALL MEAN';
 %END;
 DATA &DATA_;SET RM82152D;
 DATA RM82152D;SET RM82152D;
 ARRAY RM852X{*} &INDEPEN;

 DUM82152=&CLASS_||' ';
 DUM82152=TRIM(DUM82152);
 IF &DEPEN=. THEN DELETE;
 RM82152P=DIM(RM852X);
 DO I=1 TO RM82152P;
 IF RM852X{I}=. THEN DELETE;
 END;
 PROC SORT DATA=RM82152D;
 BY &CLASS_ &BYVAR;
 PROC MEANS N NOPRINT DATA=RM82152D;
 BY &CLASS_ &BYVAR;
 VAR &DEPEN;
 OUTPUT OUT=RM82152O N=SAMSIZE;
 DATA RM82152D;MERGE RM82152D RM82152O;
 BY &CLASS_ &BYVAR;
 RM82152C=FIRST.&CLASS_;
 RM82152S=FIRST.&BYVAR;
 RM82152S=RM82152S*SAMSIZE;

 PROC SORT DATA=&DATA_;
 BY &CLASS_ &BYVAR;
 DATA &DATA_;MERGE &DATA_ RM82152O;
 BY &CLASS_ &BYVAR;
 DUMCHARV='.';

PROC IML;
RESET STORAGE=&STORAGE;
LOAD PARMID LABELID ALPHA0 CONVERGE OLSITER MAXITER SUBITER OPTIONS METHOD
     BEST TYPE STEPS EXPAND SSMETHOD COVAR TYPE_I SIG82152 SIGSQ 
     PSI_IND PSI_STRT CSOURCE NO_C CONTRAST PQ RANK_C NCYCLE
     ESOURCE NO_E ESTIMAT RANK_E;

START GETDATA;
 RANDID={&RANDOM_};
 NAMERND={&RANDOM_}`;
 RESTRID={&RESTRCT};
   P=NCOL(PARMID);
   V=NCOL(RANDID);
   NO_RSCT=NCOL(RESTRID);
   IF RESTRID={'NONE'} THEN NO_RSCT=0;

 V082152=0;** =V IS NECESSARY TO ESTIMATE_ PSI USING METHOD OF MOMENTS ESTIMATOR;

USE &DATA_;
READ ALL VAR _CHAR_ WHERE(SAMSIZE>V082152) INTO RESULT1 (|COLNAME=OUTNAME1|);
READ ALL VAR _NUM_ WHERE(SAMSIZE>V082152) INTO RESULT2 (|COLNAME=OUTNAME2|);

USE RM82152D;
READ ALL VAR _ALL_ WHERE(SAMSIZE>V082152); /*READ ALL VARIABLES INTO OUTPRED*/
READ ALL VAR {&BYVAR} WHERE(SAMSIZE>V082152);
READ ALL VAR {&CLASS_} WHERE(SAMSIZE>V082152);
READ ALL VAR {&DEPEN} WHERE(SAMSIZE>V082152);
READ ALL VAR {&INDEPEN} WHERE(SAMSIZE>V082152) INTO INDEP (|COLNAME=INDEPNME|);
READ ALL VAR {SAMSIZE} WHERE(RM82152S>V082152) INTO SAMSIZ;
READ ALL VAR {DUM82152} WHERE(RM82152C=1) INTO DESCRIP0;

 BI_NAME=1:V;
 BI_NAME=TRIM(CHAR(BI_NAME));
 BI_NAME=CONCAT( J(1,V,"BHAT_"),BI_NAME );
 BI_NAME=COMPRESS(BI_NAME);
OUTNAME=OUTNAME1||OUTNAME2;
OUTNAME = OUTNAME||{"_PRED_" "CLM_LOW" "CLM_UP" "CLI_LOW" "CLI_UP" "PRED_SS"};
OUTNAME = OUTNAME||BI_NAME;
CREATE R082152 VAR OUTNAME; /* CREATES DATA R082152 FOR OUTPRED OPTION */
APPEND VAR OUTNAME;

FREE / PARMID LABELID ALPHA0 CONVERGE OLSITER MAXITER SUBITER OPTIONS METHOD
     BEST TYPE STEPS EXPAND SSMETHOD COVAR TYPE_I SIG82152 SIGSQ 
     PSI_IND PSI_STRT CSOURCE NO_C CONTRAST PQ RANK_C NCYCLE 
     ESOURCE NO_E ESTIMAT RANK_E 
     RANDID RESTRID P V V082152 NO_RSCT &BYVAR &DEPEN &CLASS_ 
     INDEP INDEPNME NAMERND
     SAMSIZ DESCRIP0 BI_NAME OUTNAME;


BYVAR=&BYVAR;
Y082152 = &DEPEN;
&CLASS_=DESIGN(&CLASS_);
NO_CLASS=NCOL(&CLASS_);
%INDEP;
%_X2_;
%SETDSGN;
FREE &INDEPEN;
_X82152_=DSGNALL;
Q = NCOL(_X82152_);
N = NROW(SAMSIZ);
RMIN=MIN(SAMSIZ);
RMAX=MAX(SAMSIZ);
RAVG=SUM(SAMSIZ)/N;
TOTAL=SUM(SAMSIZ);
X82152=J(N,Q,0);NCUM=0;
DO J=1 TO N;
RI=SAMSIZ[J,];
FIRST=NCUM+1;LAST=NCUM+RI;
X82152[J,]=_X82152_[FIRST,];
NCUM=NCUM+RI;
END;

/* THE FOLLOWING CODE DEFINES: SELECT, A MATRIX USED TO DEFINE Z MATRIX */ 
%INDEP;
IF NCOL(PARMID)=NCOL(UNION(RANDID,PARMID)) THEN DO;ZID=0;ZIDNAME={NULL};END;
ELSE DO;ZID=1;ZIDNAME=SETDIF(RANDID,PARMID);END;
PZ=NCOL(ZIDNAME);
SELECT=J(V,P,0);
DO I=1 TO P;
 DO J=1 TO V;
  TEMP1821=PARMID[,I];
  TEMP2821=RANDID[,J];
  IF TEMP1821=TEMP2821 THEN SELECT[J,I]=1;
 END;
END;
S082152=1;
IF EXPAND={"AVERAGE"} THEN S082152=0;
C082152=1;
IF SSMETHOD={"LS"} THEN C082152=0;

SELECT2=J(V,PZ,0);
IF ZID>0 THEN DO;
DO I=1 TO PZ;
 DO J=1 TO V;
  TEMP1821=ZIDNAME[,I];
  TEMP2821=RANDID[,J];
  IF TEMP1821=TEMP2821 THEN SELECT2[J,I]=1;
 END;
END;
END;

/* THE FOLLOWING CODE FORCES DERIVATIVE MATRIX TO ZERO FOR RESTRICTED ALPHA */
BSGN=J(P,1,1);
DO I=1 TO P;
 DO J=1 TO NO_RSCT;
  TEMP1821=PARMID[,I];
  TEMP2821=RESTRID[,J];
  IF TEMP1821=TEMP2821 THEN BSGN[I,1]=0;
 END;
END;
%_SOURCE_;
%GMANOVA;
FREE &INDEPEN;

NAMEY={"&DEPEN"};
NAMEINDE={&INDEPEN}`;
MESSAGE=&MESSAGE;
WARNING1=&WARNING;
FINISH;

RUN GETDATA;
RESET STORAGE=&STORAGE;
STORE _ALL_;
FREE /;
%MEND GETDATA;

%MACRO EXECUTE(CALLMIX) / STMT;
%GETDATA;

PROC IML;* symsize=200;
RESET STORAGE=&STORAGE ;*details; ** Put in to see details of storage;
*show space;                      ** Put in if you want to see space and memory;
*show memory;
START MODEL;
RUN MODEL1;
 F=F082152;
 X=X082152;
 _WEIGHT_=W082152;
 _PRED_=P082152;
FREE F082152 X082152 W082152 P082152;
FINISH;

START MODEL1;
 %INDEP; 
 %BETA_W;         * EVALUATE BETA AT AVERAGE RANDOM bi = 0;
 %CALLUSER;       * EVALUATE USER DEFINED FUNCTIONS;
  P082152=&FUNC;  * EVALUATE PREDICTED VALUE AT AVERGAE RANDOM bi = 0;
 %BETA_F;         * EVALUATE BETA AT BOTH CURRENT FIXED b AND RANDOM bi;
 %CALLUSER;       * EVALUATE USER FUNCTIONS AT CURRENT FIXED b AND RANDOM bi;
  W082152=&WT;    * EVALUATE WEIGHT FUNCTION AT CURRENT FIXED b AND RANDOM bi;
  F082152=&FUNC;  * EVALUATE MODEL FUNCTION AT CURRENT FIXED b AND RANDOM bi;
 %CALLDERB;       * EVALUATE DERIVATIVES AT CURRENT FIXED b AND RANDOM bi;
 %XFIX;
FREE &INDEPEN; 
FINISH;

START VARIANCE;
 %INDEP;
 %CALLUSER;
 _WEIGHT_=&WT;        * EVALUATE WEIGHT FUNCTION AT CURRENT b, bi;
 VARIANCE=1/_WEIGHT_; * EVALUATE VARIANCE FUNCTION AT CURRENT b, bi;
FREE &INDEPEN;
FINISH;


START RANDOM;      /* RANDOM MODULE DEFINES Z MATRIX FOR GENERAL MODEL */
%INDEP;
ZFIX=J(TOTAL,PZ,0);
IF ZID>0 THEN DO;
DO I=1 TO PZ;
 ZFIX[,I]=VALUE(ZIDNAME[,I]);
END;
END;
FREE &INDEPEN;
Z = ZFIX*SELECT2` + DERB*SELECT`;
FINISH;

LOAD;
NAMEB=TRIM(SUBSTR(NAMEB,1,16));
NAMEX=TRIM(SUBSTR(NAMEX,1,30));
VALUE=TRIM(SUBSTR(VALUE,1,30));
RESET NONAME ;

IF MESSAGE=1 THEN DO;
PRINT / "WARNING: THE OVERALL DESIGN MATRIX SPECIFIED BY THE ALL = STATEMENT",
        "         IS MISSING. PROCEDURE MIXNLIN REQUIRES THE ALL = STATEMENT",
        "         TO BE PRESENT EVEN IF THE INDIVIDUAL PARAMETERS, B1,B2,...",
        "         ARE ASSIGNED THEIR OWN DESIGN MATRICES. VALUES ASSIGNED TO",
        "         THE ALL = STATEMENT SHOULD RESULT IN A NONSINGULAR DESIGN ",
        "         MATRIX WHICH REFLECTS ALL OF THE REGRESSORS, INCLUDING THE",
        "         THE CLASS VARIABLE IF PRESENT, WHICH ARE DEFINED BY THE   ",
        "         INDIVIDUAL DESIGN MATRICES ASSIGNED TO B1,B2,...          ";
END;

ELSE DO;
OPTNAME={"METHOD OF ESTIMATION:","EXPAND ABOUT:","RANDOM-EFFECT ESTIMATION:",
         "STD. ERROR ESTIMATES:", "ALTERNATIVE LARGE SAMPLE F-TEST:"};
OPTVAL =METHOD//EXPAND//SSMETHOD//COVAR//TYPE;
PRINT / 'MACRO PROCEDURE MIXNLIN',,
        'VARIABLE AND CLASS LEVEL INFORMATION',,
        ' ' (|COLNAME='VARIABLE:' FORMAT=$8.|)
        NAMEY (|COLNAME='DEPENDENT' FORMAT=$8.|)
        NAMEINDE (|COLNAME='INDEPENDENT' FORMAT=$8.|)
        NAMEB (|COLNAME='PARAMETERS'|)
        NAMEX (|COLNAME='EFFECT'|)
        VALUE (|COLNAME=' '|)
        ALPHAN (|COLNAME='LEGEND' FORMAT=$6.|)
        NAMERND (|COLNAME='RANDOM EFFECTS' FORMAT=$8.|),,
        'MIXNLIN OPTIONS',,
        OPTNAME (|FORMAT=$32.|)
        OPTVAL  (|FORMAT=$10.|);

RUN GENMEM;

  RESET STORAGE=&STORAGE;
  REMOVE _ALL_ ;
END;
OPTIONS NOIMPLMAC;

%IF &_OUT_ NE %THEN %DO;
  DATA &OUTPRED1;SET R082152;
  DROP _TYPE_ _FREQ_ SAMSIZE DUMCHARV;
%END;

%MEND EXECUTE;
                                                   
